#!/usr/bin/python3
########################################################################################################################
# Copyright 2021-2022 Datum Technology Corporation
# All rights reserved.
########################################################################################################################


########################################################################################################################
# IMPORTS
########################################################################################################################
import gen.ral_typedefs
import gen.ral_templates
import re
import csv
import jinja2
import tarfile
from jinja2 import Template
from enum import Enum
import os
import sys
from datetime import date
from mio import common


########################################################################################################################
# GLOBALS
########################################################################################################################
dbg = False
out_path = ""
default_copyright_owner = "Contributors"
default_license = "SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1"
ip_name = ""
name_of_copyright_owner = ""
license = ""
csv_filepath = ""
gen_block_offset = ""
default_gen_block_offset = "0000_0000"
parameters = {}
num_files = 0


########################################################################################################################
# FUNCTIONS
########################################################################################################################
def gen_block():
    global gen_block_offset
    global ip_name
    global csv_filepath
    global name_of_copyright_owner
    global license
    global out_path
    global num_files
    gen_block_name = ip_name
    row_count = 0
    
    cfg = gen.ral_typedefs.Configuration(ip_name, 32, gen.ral_typedefs.uvm_endianness_enum.UVM_LITTLE_ENDIAN, csv_filepath, name_of_copyright_owner, license)
    block = gen.ral_typedefs.RegisterBlock(gen_block_name, gen_block_offset)
    
    with open(csv_filepath) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        
        for row in csv_reader:
            if row_count > 0:
                raw_offset   = row[0].lower()
                raw_reg_name = row[1].lower().strip()
                access_type  = row[2].upper().strip()
                field_size   = row[3].strip()
                field_lsb    = row[4].strip()
                field_name   = row[5].lower().strip()
                raw_reset    = row[6].lower().strip()
                comments     = row[7].strip()
                
                vector_regex_results = re.match(gen.ral_templates.pattern_symbol_vector, raw_reg_name)
                if vector_regex_results:
                    reg_name  = vector_regex_results.group(1)
                    reg_size  = vector_regex_results.group(2)
                    is_vector = 1
                else:
                    reg_name  = raw_reg_name
                    reg_size  = 1
                    is_vector = 0
                
                hex_regex_results = re.match(gen.ral_templates.pattern_symbol_hex, raw_offset)
                if hex_regex_results:
                    offset = hex_regex_results.group(1)
                else:
                    common.fatal("Invalid offset for field " + gen_block_name + "." + reg_name + "." + field_name + ": '" + raw_offset + "'")
                
                
                if reg_name not in block.registers:
                    block.registers[reg_name] = gen.ral_typedefs.Register(reg_name, block.name, offset, row[3], is_vector, reg_size)
                current_register = block.registers[reg_name]
                
                lsb = field_lsb
                msb = lsb + field_size
                size = field_size
                
                hex_regex_results = re.match(gen.ral_templates.pattern_symbol_hex, raw_reset)
                if hex_regex_results:
                    reset_value = int(hex_regex_results.group(1), 16)
                else:
                    common.fatal("Invalid reset value for field " + gen_block_name + "." + reg_name + "." + field_name + ": '" + raw_reset + "'")
                    
                current_register.fields[field_name] = gen.ral_typedefs.RegisterField(field_name, size, lsb, msb, reset_value, access_type, comments, reg=current_register)
                current_field = current_register.fields[field_name]
            
            row_count += 1
    
    reg_text = ""
    for reg in block.registers:
        current_reg = block.registers[reg]
        reg_text = current_reg.render(cfg)
        reg_filename = gen.ral_templates.template_sub_reg_filename.render(
            ip_name=cfg.ip_name,
            name=current_reg.name
        )
        
        if not os.path.exists(out_path):
            os.mkdir(out_path)
        reg_file = open(out_path + "/" + reg_filename, "w")
        reg_file.write(reg_text)
        reg_file.close()
        num_files += 1
    
    block_text = block.render(cfg, reg_text)
    block_filename = gen.ral_templates.template_sub_block_filename.render(
        name=cfg.ip_name
    )
    block_file = open(out_path + "/" + block_filename, "w")
    block_file.write(block_text)
    block_file.close()
    num_files += 1


def prompt_user_values():
    global out_path
    global csv_filepath
    global ip_name
    global license
    global default_license
    global name_of_copyright_owner
    global default_copyright_owner
    global parameters
    global gen_block_offset
    global default_gen_block_offset
    
    common.info("Moore.io Template Generator: UVM Register Model (v1p0)")
    common.info("***********************************************************************")
    common.info("The answers to the following questionnaire will be used to generate the code for your new UVM Register Model")
    common.info("")
    
    if out_path == "":
        out_path = common.prompt("Please enter the destination path for this new UVM Register Model (default: '.'):\n").strip()
        if out_path == "":
            out_path = "."
    
    
    name_of_copyright_owner = common.prompt("Please enter a specific name for the copyright holder or hit RETURN for the default (default is '" + default_copyright_owner + "'):\n").strip()
    if name_of_copyright_owner == "":
        name_of_copyright_owner = default_copyright_owner
    
    license = common.prompt("Please enter a usage license or hit RETURN for the default (default is '" + default_license + "'):\n").strip()
    if license == "":
        license = default_license
    
    ip_name = common.prompt("Please enter the name of the UVM Environment to which this Model belongs (ex: 'ctrl'); this name will be used for all types (ex: 'uvme_ctrl_reg_block_c'):\n").lower().strip()
    if ip_name == "":
        common.fatal("ERROR: UVM Environment name cannot be empty.  Exiting.")
    
    csv_filepath = common.prompt("Please enter the path of the input .csv file:\n").lower().strip()
    if csv_filepath == "":
        common.fatal("ERROR: Input .csv file path cannot be empty.  Exiting.")
    
    gen_block_offset = common.prompt("Please enter the base address (default: 0000_0000):\n").lower().strip()
    if gen_block_offset == "":
        gen_block_offset = default_gen_block_offset
    
    parameters = {
        "ip_name"                 : ip_name,
        "year"                    : date.today().year,
        "name_of_copyright_owner" : name_of_copyright_owner,
        "license"                 : license
    }


def print_end_message():
    global out_path
    global num_files
    common.info(str(num_files) + " files successfully generated into " + out_path)


########################################################################################################################
# ENTRY POINT
########################################################################################################################
def main():
    prompt_user_values()
    gen_block()
    print_end_message()
