# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################




from mio import common
from mio import cfg
from mio import clean
from mio import cov
from mio import dox
from mio import results

import sys
import os
import toml
from datetime import datetime
import yaml
from yaml.loader import SafeLoader
import collections
import pathlib



project_toml_file_path = "./mio.toml"
ip_paths      = {}
ip_metadata   = {}
fsoc_ip_dir_name = {}




def find_project_descriptor():
    global project_toml_file_path
    found_file = False
    current_dir = cfg.pwd
    while found_file == False:
        if not os.path.exists(current_dir):
            return False
        else:
            project_toml_file_path = os.path.join(current_dir, "mio.toml")
            project_toml_file_path = str(pathlib.Path(project_toml_file_path).resolve())
            if not os.path.exists(project_toml_file_path):
                current_dir = os.path.join(current_dir, "..")
            else:
                found_file = True
                cfg.set_project_dir(current_dir)
                common.dbg("Found Moore.io project TOML Configuration file at " + project_toml_file_path)
    return found_file


def load_tree():
    common.dbg("in load_tree()")
    builtin_toml_file_path = os.path.join(cfg.mio_data_src_dir, "mio.toml")
    user_toml_file_path = os.path.join(cfg.user_dir, "mio.toml")
    
    cfg.configuration = toml.load(builtin_toml_file_path)
    if os.path.exists(user_toml_file_path):
        merge_dict(cfg.configuration, toml.load(user_toml_file_path))
        common.dbg("Found Moore.io user TOML Configuration file at " + user_toml_file_path)
    common.merge_dict(cfg.configuration, toml.load(project_toml_file_path))
    common.dbg("Final configuration:\n" + toml.dumps(cfg.configuration))


def load_cache():
    try:
        with open(cfg.fsoc_cache_file_path, 'r') as yamlfile:
            fsoc_cache_yaml = yaml.load(yamlfile, Loader=SafeLoader)
            cfg.fsoc_cache = fsoc_cache_yaml['cores']
            common.dbg("FuseSoC cache has " + str(len(cfg.fsoc_cache)) + " IPs")
    except:
        common.warning("FuseSoC cache is damaged/malformed.  Starting fresh.")
        cfg.fsoc_cache = {}
    
    try:
        with open(cfg.ip_cache_file_path, 'r') as yamlfile:
            ip_cache_yaml = yaml.load(yamlfile, Loader=SafeLoader)
            cfg.ip_local_cache    = ip_cache_yaml['local'   ]
            cfg.ip_external_cache = ip_cache_yaml['external']
            common.dbg("IP cache has " + str(len(cfg.ip_local_cache   )) + " local IPs"   )
            common.dbg("IP cache has " + str(len(cfg.ip_external_cache)) + " external IPs")
    except:
        common.warning("IP cache is damaged/malformed.  Starting fresh.")
        cfg.ip_local_cache    = {}
        cfg.ip_external_cache = {}


def check_cache_integrity():
    core_rm_list = []
    local_ip_rm_list = []
    external_ip_rm_list = []
    for core in cfg.fsoc_cache:
        core_data = cfg.fsoc_cache[core]
        core_file_path = f"{core_data['path']}/{core_data['sname']}.core"
        if not os.path.exists(core_file_path):
            clean.clean_core_compiled_code(core_data['name'])
            common.warning(f"IP '{core_data['name']}' at location '{core_data['path']}' has been moved or removed.")
            core_rm_list.append(core_data['name'])
    for ip in cfg.ip_local_cache:
        ip_name = cfg.ip_local_cache[ip]['name']
        ip_path = cfg.ip_local_cache[ip]['path']
        if not os.path.exists(ip_path + "/ip.yml"):
            clean.clean_ip_compiled_code(ip_name)
            common.warning(f"IP '{ip_name}' at location '{ip_path}' has been moved or removed.")
            local_ip_rm_list.append(ip_name)
    for vendor in cfg.ip_external_cache:
        for ip in cfg.ip_external_cache[vendor]:
            ip_name = cfg.ip_external_cache[vendor][ip]['name']
            ip_path = cfg.ip_external_cache[vendor][ip]['path']
            if not os.path.exists(ip_path + "/ip.yml"):
                clean.clean_ip_compiled_code(ip_name)
                common.warning(f"IP '{ip_name}' at location '{ip_path}' has been moved or removed.")
                local_ip_rm_list.append({'vendor': vendor, 'name': ip_name})
    
    for core in core_rm_list:
        cfg.fsoc_cache.pop(core, None)
    for ip in local_ip_rm_list:
        cfg.ip_local_cache.pop(ip, None)
    for pair in external_ip_rm_list:
        cfg.ip_external_cache[pair['vendor']][pair['name']].pop(ip, None)


def load_configuration():
    cfg.project_name      = cfg.configuration.get("project", {}).get("name")
    cfg.sim_dir           = os.path.join(cfg.project_dir, cfg.configuration.get("simulation", {}).get("root-path"))
    cfg.sim_results_dir   = os.path.join(cfg.sim_dir    , cfg.configuration.get("simulation", {}).get("results-dir"))
    cfg.regr_results_dir  = os.path.join(cfg.sim_dir    , cfg.configuration.get("simulation", {}).get("regressions-dir"))
    cfg.ip_paths          = cfg.configuration.get("ip", {}).get("paths")
    cfg.test_results_path_template = cfg.configuration.get("simulation", {}).get("test-result-path-template")


def scan_and_load_ip_metadata():
    load_cache()
    check_cache_integrity()
    for ip_path in cfg.ip_paths:
        full_ip_path = os.path.join(cfg.project_dir, ip_path)
        find_fsoc_cores(full_ip_path, ip_path)
        find_local_ip_files(full_ip_path)
    find_external_ip_files(cfg.dependencies_path   )
    find_external_ip_files(cfg.user_global_ips_path)


def get_ip(name, no_fatal=False):
    if name in cfg.ip_local_cache:
        return cfg.ip_local_cache[name]
    else:
        return get_external_ip(name, no_fatal)


def get_external_ip(name, no_fatal=False):
    found_ip = False
    for vendor in cfg.ip_external_cache:
        if name in cfg.ip_external_cache[vendor]:
            found_ip = True
            return cfg.ip_external_cache[vendor][name]
    if found_ip == False and no_fatal == False:
        common.fatal(f"Could not find IP '{name}'.")


def check_ip(name):
    if name in cfg.ip_local_cache:
        return True
    else:
        ip = get_external_ip(name)
        if ip != None:
            return True


def check_target_ip():
    if cfg.target_ip_name != "":
        cfg.target_ip = get_ip(cfg.target_ip_name)
        if cfg.target_ip_name in cfg.ip_local_cache:
            cfg.target_ip_is_local = True
        else:
            cfg.target_ip_is_local = False
        check_ip(cfg.target_ip_name)
    else:
        common.fatal("No IP specified!")


def find_fsoc_cores(path, ip_path):
    add_core = False
    for dirpath, dirnames, filenames in os.walk(path):
        for dir in dirnames:
            current_dir_path       = os.path.join(path            , dir)
            current_core_file_path = os.path.join(current_dir_path, dir + ".core")
            if os.path.exists(current_core_file_path):
                common.dbg("Found FuseSoC IP! name='" + dir + "', path='" + current_dir_path + "'")
                try:
                    with open(current_core_file_path, 'r') as yamlfile:
                        core_yaml = yaml.load(yamlfile, Loader=SafeLoader)
                        core_name = core_yaml["name"].strip()
                except Exception as e:
                    common.warning("FuseSoC Core file " + current_core_file_path + " is malformed and will be ignored: " + str(e))
                    continue
                hash = common.calc_file_hash(current_core_file_path)
                if core_name not in cfg.fsoc_cache:
                    add_core = True
                else:
                    if hash != cfg.fsoc_cache[core_name]['hash']:
                        add_core = True
                if add_core:
                    cfg.fsoc_cache[core_name] = {}
                    cfg.fsoc_cache[core_name]['ip_path' ] = ip_path
                    cfg.fsoc_cache[core_name]['sname'] = dir
                    cfg.fsoc_cache[core_name]['name'] = core_name
                    cfg.fsoc_cache[core_name]['path'] = current_dir_path
                    cfg.fsoc_cache[core_name]['data'] = core_yaml
                    cfg.fsoc_cache[core_name]['hash'] = hash
                    cfg.fsoc_cache[core_name]['installed'] = False
                    common.dbg("Added FuseSoc core " + core_name + " to cache")


def find_local_ip_files(path):
    add_ip = False
    for dirpath, dirnames, filenames in os.walk(path):
        for dir in dirnames:
            current_dir_path     = os.path.join(path            , dir)
            current_ip_file_path = os.path.join(current_dir_path, "ip.yml")
            if os.path.exists(current_ip_file_path):
                common.dbg("Found IP! name='" + dir + "', path='" + current_dir_path + "'")
                try:
                    with open(current_ip_file_path, 'r') as yamlfile:
                        ip_yaml = yaml.load(yamlfile, Loader=SafeLoader)
                        ip_name = ip_yaml["ip"]["name"].strip()
                except Exception as e:
                    common.warning("IP file " + current_ip_file_path + " is malformed and will be ignored: " + str(e))
                    continue
                if ip_name not in cfg.ip_local_cache:
                    add_ip = True
                if add_ip:
                    cfg.ip_local_cache[ip_name] = {}
                    cfg.ip_local_cache[ip_name]['name'] = ip_name
                    cfg.ip_local_cache[ip_name]['path'] = current_dir_path
                    cfg.ip_local_cache[ip_name]['local'] = True
                    cfg.ip_local_cache[ip_name]['installed'] = False
                    common.dbg(f"Added IP {ip_name} local IP cache")
                cfg.ip_local_cache[ip_name]['data'] = ip_yaml


def find_external_ip_files(path):
    add_ip = False
    for dirpath, dirnames, filenames in os.walk(path):
        for dir in dirnames:
            current_dir_path     = os.path.join(path            , dir)
            current_ip_file_path = os.path.join(current_dir_path, "ip.yml")
            if os.path.exists(current_ip_file_path):
                common.dbg("Found IP! name='" + dir + "', path='" + current_dir_path + "'")
                try:
                    with open(current_ip_file_path, 'r') as yamlfile:
                        ip_yaml = yaml.load(yamlfile, Loader=SafeLoader)
                        ip_name = ip_yaml["ip"]["name"].strip()
                except Exception as e:
                    common.warning(f"IP file {current_ip_file_path} is malformed and will be ignored: {str(e)}")
                    continue
                hash = common.calc_file_hash(current_ip_file_path)
                #ip_vendor = ip_yaml['ip']['vendor']
                ip_vendor = "datum"  # TODO Add vendor tags to all IPs in marketplace
                if ip_vendor not in cfg.ip_external_cache:
                    cfg.ip_external_cache[ip_vendor] = {}
                    add_ip = True
                else:
                    if ip_name not in cfg.ip_external_cache[ip_vendor]:
                        add_ip = True
                    else:
                        if hash != cfg.ip_external_cache[ip_vendor][ip_name]['hash']:
                            add_ip = True
                if add_ip:
                    cfg.ip_external_cache[ip_vendor][ip_name] = {}
                    cfg.ip_external_cache[ip_vendor][ip_name]['name'] = ip_name
                    cfg.ip_external_cache[ip_vendor][ip_name]['path'] = current_dir_path
                    cfg.ip_external_cache[ip_vendor][ip_name]['data'] = ip_yaml
                    cfg.ip_external_cache[ip_vendor][ip_name]['local'] = False
                    cfg.ip_external_cache[ip_vendor][ip_name]['compiled'] = {}
                    cfg.ip_external_cache[ip_vendor][ip_name]['compiled']['vivado' ] = False
                    cfg.ip_external_cache[ip_vendor][ip_name]['compiled']['metrics'] = False
                    cfg.ip_external_cache[ip_vendor][ip_name]['compiled']['vcs'    ] = False
                    cfg.ip_external_cache[ip_vendor][ip_name]['compiled']['xcelium'] = False
                    cfg.ip_external_cache[ip_vendor][ip_name]['compiled']['questa' ] = False
                    cfg.ip_external_cache[ip_vendor][ip_name]['compiled']['riviera'] = False
                    cfg.ip_external_cache[ip_vendor][ip_name]['hash'] = hash
                    common.dbg(f"Added IP '{ip_vendor}' {ip_name}' to local IP cache")


def write_caches_to_disk():
    try:
        with open(cfg.ip_cache_file_path, 'w') as yaml_file_write:
            ip_cache = {
                "local"    : cfg.ip_local_cache,
                "external" : cfg.ip_external_cache
            }
            yaml.dump(ip_cache, yaml_file_write)
        with open(cfg.fsoc_cache_file_path, 'w') as yaml_file_write:
            yaml.dump({ "cores" : cfg.fsoc_cache }, yaml_file_write)
        with open(cfg.job_history_file_path, 'w') as yamlfile:
            yaml.dump({"history" : cfg.job_history}, yamlfile)
    except Exception as e:
        print("\033[31m\033[1m[mio-fatal] Could not write caches to disk \033[0m: " + str(e))
        sys.exit(0)



