# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import common
from mio import cache
from mio import cfg
from jinja2 import Template
from fusesoc import main as fsoc
import fusesoc

import pathlib
import argparse
import os
import subprocess
import re
import yaml
from yaml.loader import SafeLoader
from threading import BoundedSemaphore


vivado_default_compilation_args  = ["--incr", "-sv"]
metrics_default_compilation_args = []
vcs_default_compilation_args     = []
xcelium_default_compilation_args = []
questa_default_compilation_args  = []
riviera_default_compilation_args = []

vivado_project_default_vlog_compilation_args = ["--relax"]
vivado_project_default_vhdl_compilation_args = ["--relax"]

vivado_default_elaboration_args  = ["--incr", "-relax", "--O0", "-v 0", "-timescale 1ns/1ps", "-dup_entity_as_module"]
metrics_default_elaboration_args = []
vcs_default_elaboration_args     = []
xcelium_default_elaboration_args = []
questa_default_elaboration_args  = []
riviera_default_elaboration_args = []

vivado_default_simulation_args  = ["--stats"]
metrics_default_simulation_args = []
vcs_default_simulation_args     = []
xcelium_default_simulation_args = []
questa_default_simulation_args  = []
riviera_default_simulation_args = []

vivado_cmp_log_error_regexes  = ["ERROR:"]
metrics_cmp_log_error_regexes = []
vcs_cmp_log_error_regexes     = []
xcelium_cmp_log_error_regexes = []
questa_cmp_log_error_regexes  = []
riviera_cmp_log_error_regexes = []

vivado_elab_log_error_regexes  = ["ERROR:"]
metrics_elab_log_error_regexes = []
vcs_elab_log_error_regexes     = []
xcelium_elab_log_error_regexes = []
questa_elab_log_error_regexes  = []
riviera_elab_log_error_regexes = []

sem = BoundedSemaphore(1)


def compile_fsoc_core(flist_path, core, defines, sim_args):
    core_name = core['name']
    timestamp_start = common.timestamp()
    var_name = core['sname'].replace("-", "_").upper() + "_SRC_PATH"
    os.environ[var_name] = core['path']
    common.dbg(f"FuseSoC env var '{var_name}'='{core['path']}'")
    log_file_path = compile_flist(core['sname'], flist_path, [], defines, sim_args, local=True)
    timestamp_end = common.timestamp()
    errors = scan_cmp_log_file_for_errors(log_file_path, sim_args)
    if len(errors):
        common.error("Errors during compilation of FuseSoC core '" + core_name + "':")
        for error in errors:
            common.error("  " + error)
        common.fatal("Stopping due to compilation.")
    log_cmp_history_fsoc(core, log_file_path, sim_args, timestamp_start, timestamp_end)


def compile_ip(ip_yml, defines, sim_args, local):
    ip_name = ip_yml['name']
    flist_path = get_ip_flist_path(ip_yml, sim_args)
    deps_list = get_dep_list(ip_yml, sim_args)
    timestamp_start = common.timestamp()
    os.environ[ip_name.upper() + '_SRC_PATH'] = ip_yml['path'] + "/" + ip_yml['data']['structure']['src-path']
    log_file_path = compile_flist(ip_name, flist_path, deps_list, defines, sim_args, local)
    timestamp_end = common.timestamp()
    errors = scan_cmp_log_file_for_errors(log_file_path, sim_args)
    if len(errors):
        common.error("Errors during compilation of IP '" + ip_name + "':")
        for error in errors:
            common.error("  " + error)
        common.fatal("Stopping due to compilation errors. Full log: " + log_file_path)
    log_cmp_history_ip(ip_yml, log_file_path, sim_args, timestamp_start, timestamp_end)
    if ip_name not in cfg.ip_local_cache:
        ip_yml['compiled'][sim_args['simulator']] = True
    return log_file_path


def compile_vivado_project(ip_yml, defines, sim_args, local):
    ip_name = ip_yml['name']
    dep_list = get_dep_list(ip_yml, sim_args)
    timestamp_start = common.timestamp()
    log_file_paths = do_compile_vivado_project(ip_yml, dep_list, defines, sim_args, local)
    timestamp_end = common.timestamp()
    errors = scan_cmp_log_file_for_errors(log_file_paths[0], sim_args)
    errors.append(scan_cmp_log_file_for_errors(log_file_paths[1], sim_args))
    if len(errors):
        common.error("Errors during compilation of Vivado Project IP '" + ip_name + "':")
        for error in errors:
            common.error("  " + error)
        common.fatal("Stopping due to compilation errors. Full log: " + log_file_path)
    log_cmp_history_vivado_project(ip_yml, log_file_paths[0], log_file_paths[1], sim_args, timestamp_start, timestamp_end)
    if ip_name not in cfg.ip_local_cache:
        ip_yml['compiled'][sim_args['simulator']] = True
    return log_file_paths


def elaborate(ip_yml, defines, sim_args, is_regression=False, regression_name="", regression_timestamp=""):
    ip_name = ip_yml['name']
    deps_list = get_dep_list(ip_yml, sim_args)
    timestamp_start = common.timestamp()
    if is_regression:
        common.create_dir(cfg.sim_output_dir + "/" + sim_args['simulator'] + "/regressions/" + cfg.target_ip_name + "_" + regression_name)
        common.create_dir(cfg.sim_output_dir + "/" + sim_args['simulator'] + "/regressions/" + cfg.target_ip_name + "_" + regression_name + "/" + regression_timestamp)
        elab_out =        cfg.sim_output_dir + "/" + sim_args['simulator'] + "/regressions/" + cfg.target_ip_name + "_" + regression_name + "/" + regression_timestamp
        common.create_dir(elab_out)
    else:
        elab_out = cfg.sim_output_dir + "/" + sim_args['simulator'] + "/sim"
    log_file_path = do_elaborate(ip_yml, deps_list, defines, sim_args, elab_out)
    timestamp_end = common.timestamp()
    errors = scan_elab_log_file_for_errors(log_file_path, sim_args)
    if len(errors):
        common.error("Errors during elaboration of IP '" + ip_name + "':")
        for error in errors:
            common.error("  " + error)
        common.fatal("Stopping due to elaboration errors. Full log: " + log_file_path)
    log_elab_history(ip_yml, log_file_path, sim_args, timestamp_start, timestamp_end, is_regression, regression_name="", regression_timestamp="")
    return elab_out


def simulate(ip_yml, plus_args, sim_args, is_regression=False, regression_name="", regression_timestamp=""):
    global sem
    ip_name = ip_yml['name']
    start = common.timestamp()
    log_sim_start_history(ip_yml, sim_args, start, is_regression, regression_name, regression_timestamp)
    if is_regression:
        sim_out = cfg.sim_output_dir + "/" + sim_args['simulator'] + "/regressions/" + cfg.target_ip_name + "_" + regression_name + "/" + regression_timestamp
    else:
        sim_out = cfg.sim_output_dir + "/" + sim_args['simulator'] + "/sim"
    log_file_path, results_path, sim_args = do_simulate(ip_yml, plus_args, sim_args, sim_out, is_regression, regression_name, regression_timestamp)
    sem.acquire()
    log_sim_end_history(ip_yml, log_file_path, results_path, sim_args, start, common.timestamp(), is_regression, regression_name, regression_timestamp)
    sem.release()
    return sim_args


def compile_flist(name, flist_path, deps, defines, cli_args, local):
    arg_list = []
    def_list      = convert_defines(defines, cli_args)
    cmp_args_list = convert_compilation_args(cli_args)
    deps_list     = convert_deps_to_args(deps)
    incdir_list   = get_incdir_list(deps, cli_args)
    sim_out = cfg.sim_output_dir + "/" + cli_args['simulator']
    if local:
        cmp_out = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/local/" + name
    else:
        cmp_out = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/external/" + name
    
    if cli_args['simulator'] == "vivado":
        compilation_log_path = cfg.sim_dir + "/cmp/" + name + ".viv.cmp.log"
        arg_list += vivado_default_compilation_args
        arg_list += incdir_list
        arg_list.append("-f " + flist_path)
        arg_list += def_list
        arg_list += cmp_args_list
        arg_list.append("-L uvm")
        arg_list += deps_list
        arg_list.append("--work " + name + "=" + cmp_out)
        arg_list.append("--log "  + compilation_log_path)
        launch_eda_bin(cfg.vivado_home + "/xvlog", arg_list, wd=sim_out)
    elif cli_args['simulator'] == "vcs":
        compilation_log_path = cfg.sim_dir + "/cmp/" + name + ".vcs.cmp.log"
        arg_list += vcs_default_compilation_args
        arg_list.append("-f " + flist_path)
        # TODO Add compilation output argument for vcs
        launch_eda_bin(cfg.vcs_home + "/vcs", arg_list, wd=sim_out)
    elif cli_args['simulator'] == "metrics":
        compilation_log_path = cfg.sim_dir + "/cmp/" + name + ".mtr.cmp.log"
        arg_list += metrics_default_compilation_args
        arg_list.append("-f " + flist_path)
        # TODO Add compilation output argument for metrics
        launch_eda_bin(cfg.metrics_home + "/dsim", arg_list, wd=sim_out)
    elif cli_args['simulator'] == "xcelium":
        compilation_log_path = cfg.sim_dir + "/cmp/" + name + ".xcl.cmp.log"
        arg_list += xcelium_default_compilation_args
        arg_list.append("-f " + flist_path)
        # TODO Add compilation output argument for nc
        launch_eda_bin(cfg.nc_home + "/xrun", arg_list, wd=sim_out)
    elif cli_args['simulator'] == "questa":
        compilation_log_path = cfg.sim_dir + "/cmp/" + name + ".qst.cmp.log"
        arg_list += questa_default_compilation_args
        arg_list.append("-f " + flist_path)
        # TODO Add compilation output argument for questa
        launch_eda_bin(cfg.questa_home + "/vlog", arg_list, wd=sim_out)
    elif cli_args['simulator'] == "riviera":
        compilation_log_path = cfg.sim_dir + "/cmp/" + name + ".riv.cmp.log"
        arg_list += riviera_default_compilation_args
        arg_list.append("-f " + flist_path)
        # TODO Add compilation output argument for riviera
        launch_eda_bin(cfg.riviera_home + "/vlog", arg_list, wd=sim_out)
    
    return compilation_log_path


def do_compile_vivado_project(ip_yml, deps, defines, cli_args, local):
    ip_name = ip_yml['name']
    if cli_args['simulator'] == "vivado":
        if local:
            cmp_out = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/local/" + ip_name
        else:
            cmp_out = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/external/" + ip_name
        sim_out = cfg.sim_output_dir + "/" + cli_args['simulator']
        vlog_compilation_log_path = cfg.sim_dir + "/cmp/" + ip_name + ".viv.xvlog.cmp.log"
        vhdl_compilation_log_path = cfg.sim_dir + "/cmp/" + ip_name + ".viv.xvhdl.cmp.log"
        vlog_arg_list = ["-prj "  + ip_yml['data']['hdl-src']['vlog']]
        vlog_arg_list.append("--work " + ip_name + "=" + cmp_out)
        vlog_arg_list += vivado_project_default_vlog_compilation_args
        vlog_arg_list += convert_defines(defines)
        vlog_arg_list += convert_compilation_args(cli_args)
        vlog_arg_list += convert_deps_to_args(deps)
        vlog_arg_list.append("--log " + vlog_compilation_log_path)
        vhdl_arg_list = ["-prj " + ip_yml['data']['hdl-src']['vhdl']]
        vhdl_arg_list.append("--work " + ip_name + "=" + cmp_out)
        vhdl_arg_list += vivado_project_default_vhdl_compilation_args
        vlog_arg_list += convert_defines(defines)
        vlog_arg_list += convert_compilation_args(cli_args)
        vlog_arg_list += convert_deps_to_args(deps)
        vhdl_arg_list.append("--log " + vhdl_compilation_log_path)
        launch_eda_bin(cfg.vivado_home + "/xvlog", vlog_arg_list, wd=sim_out)
        launch_eda_bin(cfg.vivado_home + "/xvhdl", vhdl_arg_list, wd=sim_out)
    else:
        common.fatal("Vivado Project IP are not compatible with simulator '" + cli_args['simulator'] + "'.")
    
    return [vlog_compilation_log_path, vhdl_compilation_log_path]


def do_elaborate(ip_yml, deps, defines, cli_args, wd):
    ip_name = ip_yml['name']
    arg_list = []
    
    if 'dut' in ip_yml['data']:
        arg_list += dut_elab_to_arg_list(ip_yml['data'], cli_args)
    if ip_name in cfg.ip_local_cache:
        cmp_out = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/local/" + ip_name
    else:
        cmp_out = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/external/" + ip_name
    
    top_constructs = ip_yml['data']['hdl-src']['top-constructs']
    def_list  = convert_defines(defines, cli_args)
    elab_list = convert_elaboration_args(cli_args)
    deps_list = convert_deps_to_args(deps)
    
    if cli_args['simulator'] == "vivado":
        elaboration_log_path = cfg.sim_dir + "/elab/" + ip_name + ".viv.elab.log"
        arg_list += def_list
        arg_list += elab_list
        arg_list += deps_list
        arg_list += vivado_default_elaboration_args
        arg_list.append("--log "  + elaboration_log_path)
        arg_list.append("-s "     + ip_name)
        arg_list.append("-L "     + ip_name + "=" + cmp_out)
        for construct in top_constructs:
            if "." in construct:
                arg_list.append(construct)
            else:
                arg_list.append(ip_name + "." + construct)
        launch_eda_bin(cfg.vivado_home + "/xelab", arg_list, wd)
    elif cli_args['simulator'] == "vcs":
        elaboration_log_path = cfg.sim_dir + "/elab/" + ip_name + ".vcs.elab.log"
        arg_list += vcs_default_elaboration_args
        # TODO Add elaboration output argument for vcs
        launch_eda_bin(cfg.vcs_home + "/vcs", arg_list, wd)
    elif cli_args['simulator'] == "metrics":
        elaboration_log_path = cfg.sim_dir + "/elab/" + ip_name + ".mtr.elab.log"
        arg_list += metrics_default_elaboration_args
        # TODO Add elaboration output argument for metrics
        launch_eda_bin(cfg.metrics_home + "/dsim", arg_list, wd)
    elif cli_args['simulator'] == "xcelium":
        elaboration_log_path = cfg.sim_dir + "/elab/" + ip_name + ".xcl.elab.log"
        arg_list += xcelium_default_elaboration_args
        # TODO Add elaboration output argument for nc
        launch_eda_bin(cfg.nc_home + "/xrun", arg_list, wd)
    elif cli_args['simulator'] == "questa":
        elaboration_log_path = cfg.sim_dir + "/elab/" + ip_name + ".qst.elab.log"
        arg_list += questa_default_elaboration_args
        # TODO Add elaboration output argument for questa
        launch_eda_bin(cfg.questa_home + "/vopt", arg_list, wd)
    elif cli_args['simulator'] == "riviera":
        elaboration_log_path = cfg.sim_dir + "/elab/" + ip_name + ".riv.elab.log"
        arg_list += riviera_default_elaboration_args
        # TODO Add elaboration output argument for riviera
        launch_eda_bin(cfg.riviera_home + "/vlog", arg_list, wd)
    
    return elaboration_log_path


def do_simulate(ip_yml, plus_args, cli_args, wd, is_regression, regression_name, regression_timestamp):
    global sem
    sem.acquire()
    ip_name = ip_yml['name']
    ip_path = ip_yml['path']
    if (len(plus_args) > 0):
        args_present = True
    else:
        args_present = False
    arg_list = []
    test_template = Template(ip_yml['data']['hdl-src']['test-name-template'])
    test_result_dir_template = Template(cfg.test_results_path_template)
    test = cli_args['test_name']
    test_name = test_template.render(name=test)
    test_result_dir = test_result_dir_template.render(ip_name=ip_name, test_name=test, seed=cli_args['seed'], args=plus_args_list_to_str_list(plus_args), args_present=args_present)
    plus_args["UVM_TESTNAME"] = test_name
    
    if is_regression:
        tests_results_path = cfg.sim_dir + "/regressions/" + cfg.target_ip_name + "_" + regression_name + "/" + regression_timestamp + "/" + test_result_dir
    else:
        tests_results_path = cfg.sim_results_dir + "/" + test_result_dir
    
    output = not is_regression
    
    simulation_log_path = tests_results_path + "/sim.log"
    cov_path = tests_results_path + "/cov"
    common.create_dir(tests_results_path)
    common.create_dir(tests_results_path + "/trn_log")
    cli_args['results_path'] = tests_results_path
    if cli_args['waves']:
        if cli_args['simulator'] == "vivado":
            wave_capture_script_path = tests_results_path + "/waves.viv.tcl"
            waves_path = tests_results_path + "/waves.wdb"
            arg_list.append("--wdb "      + waves_path              )
            arg_list.append("--tclbatch " + wave_capture_script_path)
            if not os.path.exists(wave_capture_script_path):
                try:
                    common.create_file(wave_capture_script_path)
                    f = open(wave_capture_script_path, "w")
                    f.write("log_wave -recursive * \n")
                    f.write("run -all \n")
                    f.write("quit \n")
                    f.close()
                except Exception as e:
                    common.fatal("Could not create wave capture script at " + wave_capture_script_path + ": " + str(e))
        elif cli_args['simulator'] == "vcs":
            # TODO Implement waves file for vcs
            pass
        elif cli_args['simulator'] == "metrics":
            # TODO Implement waves file for metrics
            pass
        elif cli_args['simulator'] == "xcelium":
            # TODO Implement waves file for xcelium
            pass
        elif cli_args['simulator'] == "questa":
            # TODO Implement waves file for questa
            pass
        elif cli_args['simulator'] == "riviera":
            # TODO Implement waves file for riviera
            pass
    
    if cli_args['cov']:
        common.create_dir(cov_path)
        if cli_args['simulator'] == "vivado":
            arg_list.append("-cov_db_name " + test_name)
            arg_list.append("-cov_db_dir "  + cov_path )
        elif cli_args['simulator'] == "vcs":
            # TODO Implement coverage for vcs
            pass
        elif cli_args['simulator'] == "metrics":
            # TODO Implement coverage for metrics
            pass
        elif cli_args['simulator'] == "xcelium":
            # TODO Implement coverage for xcelium
            pass
        elif cli_args['simulator'] == "questa":
            # TODO Implement coverage for questa
            pass
        elif cli_args['simulator'] == "riviera":
            # TODO Implement coverage for riviera
            pass
    
    plus_args["UVM_NO_RELNOTES"                ] = ""
    plus_args["SIM_DIR_RESULTS"                ] = cfg.sim_results_dir
    plus_args["UVM_VERBOSITY"                  ] = "UVM_" + cli_args['verbosity'].upper()
    plus_args["UVML_FILE_BASE_DIR_SIM"         ] = cfg.sim_dir
    plus_args["UVML_FILE_BASE_DIR_TB"          ] = ip_path + "/" + ip_yml['data']['structure']['src-path']
    plus_args["UVML_FILE_BASE_DIR_TESTS"       ] = ip_path + "/" + ip_yml['data']['structure']['src-path'] + "/" + ip_yml['data']['hdl-src']['tests-path']
    plus_args["UVML_FILE_BASE_DIR_TEST_RESULTS"] = tests_results_path
    plus_args_list = convert_plus_args(plus_args, cli_args)
    if cli_args['simulator'] == "vivado":
        arg_list += plus_args_list
        arg_list += vivado_default_simulation_args
        arg_list.append("--log " + simulation_log_path)
        if cli_args['gui']:
            arg_list.append("--gui")
        else:
            if not cli_args['waves']:
                arg_list.append("--runall")
                arg_list.append("--onerror quit")
        if not cli_args['cov']:
            arg_list.append("-ignore_coverage")
        arg_list.append(ip_name)
        arg_list.append("-sv_seed " + str(cli_args['seed']))
        sem.release()
        launch_eda_bin(cfg.vivado_home + "/xsim", arg_list, wd, output=output)
    elif cli_args['simulator'] == "vcs":
        arg_list += vcs_default_simulation_args
        # TODO Add simulation output argument for vcs
        sem.release()
        launch_eda_bin(cfg.vcs_home + "/simv", arg_list, wd, output=output)
    elif cli_args['simulator'] == "metrics":
        arg_list += metrics_default_simulation_args
        # TODO Add simulation output argument for metrics
        sem.release()
        launch_eda_bin(cfg.metrics_home + "/dsim", arg_list, wd, output=output)
    elif cli_args['simulator'] == "xcelium":
        arg_list += xcelium_default_simulation_args
        # TODO Add simulation output argument for nc
        sem.release()
        launch_eda_bin(cfg.nc_home + "/xrun", arg_list, wd, output=output)
    elif cli_args['simulator'] == "questa":
        arg_list += questa_default_simulation_args
        # TODO Add simulation output argument for questa
        sem.release()
        launch_eda_bin(cfg.questa_home + "/vsim", arg_list, wd, output=output)
    elif cli_args['simulator'] == "riviera":
        arg_list += riviera_default_simulation_args
        # TODO Add simulation output argument for riviera
        sem.release()
        launch_eda_bin(cfg.riviera_home + "/vsim", arg_list, wd, output=output)
    
    return [ simulation_log_path, tests_results_path, cli_args]


def dut_elab_to_arg_list(ip_yml, cli_args):
    args = []
    dut_ip_type = ip_yml['dut']['ip-type'].lower().strip()
    if (dut_ip_type == "mio"):
        dut_ip_name = ip_yml['dut']['name']
        dut_ip_yml  = cache.get_ip(dut_ip_name)
        dut_ip_sub_type = ip_yml['dut']['sub-type'].lower().strip()
        if (dut_ip_sub_type == "vivado"):
            if cli_args['simulator'] != "vivado":
                common.fatal("Vivado Projects are only compatible with the Vivado simulator.")
            prj_lib_name       = dut_ip_yml['data']['hdl-src']['lib-name']
            prj_libs           = dut_ip_yml['data']['hdl-src']['xilinx-libs']
            prj_top_constructs = dut_ip_yml['data']['hdl-src']['top-constructs']
            for construct in prj_top_constructs:
                args.append(construct)
            for lib in prj_libs:
                args.append("-L " + lib)
        elif (dut_ip_sub_type == "flist"):
            pass
        else:
            common.fatal("Elaboration of non-Vivado Project IP DUTs not yet suppoted.")
    elif (dut_ip_type == "fsoc"):
        fsoc_name  = ip_yml['dut']['name']
        fsoc_fname = ip_yml['dut']['full-name']
        file_path_partial_name = re.sub(r':', '_', fsoc_fname)
        eda_file_dir = cfg.fsoc_dir + "/" + fsoc_name + "/sim-xsim"
        eda_file_path = eda_file_dir + "/" + file_path_partial_name + "_0.eda.yml"
        try:
            if os.path.exists(eda_file_path):
                with open(eda_file_path, 'r') as edafile:
                    eda_yaml = yaml.load(edafile, Loader=SafeLoader)
                    if eda_yaml:
                        elab_options = eda_yaml['tool_options']['xsim']['xelab_options']
                        args.append("-L " + fsoc_name + "=" + cfg.sim_output_dir + "/" + cli_args['simulator'] + "/local/" + fsoc_name)
                    else:
                        common.fatal("ERROR: Unable to parse FuseSoC output " + eda_file_path)
            else:
                common.fatal("ERROR: Unable to parse FuseSoC output " + eda_file_path)
        except Exception as e:
            common.fatal("ERROR: Unable to find FuseSoC core output for '" + dut_ip_name + "': " + str(e))
    else:
        common.fatal("DUT type not supported: " + dut_ip_type)
    common.dbg("DUT args list: " + str(args))
    return args


def get_ip_flist_path(ip_yml, cli_args):
    if cli_args['simulator'] == "vivado":
        flist_path = ip_yml['path'] + "/" + ip_yml['data']['structure']['src-path'] + "/" + ip_yml['data']['hdl-src']['flist']['vivado']
    elif cli_args['simulator'] == "metrics":
        flist_path = ip_yml['path'] + "/" + ip_yml['data']['structure']['src-path'] + "/" + ip_yml['data']['hdl-src']['flist']['metrics']
    elif cli_args['simulator'] == "vcs":
        flist_path = ip_yml['path'] + "/" + ip_yml['data']['structure']['src-path'] + "/" + ip_yml['data']['hdl-src']['flist']['vcs']
    elif cli_args['simulator'] == "xcelium":
        flist_path = ip_yml['path'] + "/" + ip_yml['data']['structure']['src-path'] + "/" + ip_yml['data']['hdl-src']['flist']['xcelium']
    elif cli_args['simulator'] == "questa":
        flist_path = ip_yml['path'] + "/" + ip_yml['data']['structure']['src-path'] + "/" + ip_yml['data']['hdl-src']['flist']['questa']
    elif cli_args['simulator'] == "riviera":
        flist_path = ip_yml['path'] + "/" + ip_yml['data']['structure']['src-path'] + "/" + ip_yml['data']['hdl-src']['flist']['riviera']
    return flist_path


def convert_defines(defines, cli_args):
    args = []
    for define in defines:
        if cli_args['simulator'] == 'vivado':
            if defines[define] != "":
                args.append("--define " + define + "=" + defines[define])
            else:
                args.append("--define " + define)
        else:
            if defines[define] != "":
                args.append("+define+" + define + "=" + defines[define])
            else:
                args.append("+define+" + define)
    return args


def convert_plus_args(plus_args, cli_args):
    args = []
    for arg in plus_args:
        if cli_args['simulator'] == 'vivado':
            if plus_args[arg] != "":
                args.append("-testplusarg \"" + arg + "=" + plus_args[arg] + "\"")
            else:
                args.append("-testplusarg \"" + arg + "\"")
        else:
            if plus_args[arg] != "":
                args.append("+" + arg + "=" + plus_args[arg])
            else:
                args.append("+" + arg)
    return args


def convert_deps_to_args(deps):
    args = []
    for dep in deps:
        if deps[dep]['name'] != "uvm":
            args.append("-L " + dep + "=" + deps[dep]['cmp_path'])
    return args


def get_dep_list(ip_yml, cli_args):
    dep_list = {}
    
    if 'dut' in ip_yml['data']:
        if ip_yml['data']['dut']['ip-type'] != "fsoc":
            ip_name = ip_yml['data']['dut']['name']
            ip = cache.get_ip(ip_name)
            dep_list[ip_name] = {}
            dep_list[ip_name]['name'] = ip_name
            dep_list[ip_name]['src_paths'] = []
            for dir in ip['data']['hdl-src']['directories']:
                if dir == ".":
                    dir = ""
                dep_list[ip_name]['src_paths'].append(ip['path'] + "/" + ip['data']['structure']['src-path'] + "/" + dir)
            if ip_name in cfg.ip_local_cache:
                dep_list[ip_name]['cmp_path']  = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/local/" + ip_name
            else:
                dep_list[ip_name]['cmp_path']  = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/external/" + ip_name
    if ip_yml['data']['dependencies'] == None:
        return dep_list
    for dep in ip_yml['data']['dependencies']:
        if dep['name'] == "uvm":
            continue
        if dep['name'] in cfg.ip_local_cache:
            ip = cache.get_ip(dep['name'])
            dep_list[dep['name']] = {}
            dep_list[dep['name']]['name'] = dep['name']
            dep_list[dep['name']]['src_paths'] = []
            for dir in ip['data']['hdl-src']['directories']:
                if dir == ".":
                    dir = ""
                dep_list[dep['name']]['src_paths'].append(ip['path'] + "/" + ip['data']['structure']['src-path'] + "/" + dir)
            dep_list[dep['name']]['cmp_path'] = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/local/" + dep['name']
        else:
            ip = cache.get_ip(dep['name'])
            dep_list[dep['name']] = {}
            dep_list[dep['name']]['name'] = dep['name']
            dep_list[dep['name']]['src_paths'] = []
            for dir in ip['data']['hdl-src']['directories']:
                if dir == ".":
                    dir = ""
                dep_list[dep['name']]['src_paths'].append(ip['path'] + "/" + ip['data']['structure']['src-path'] + "/" + dir)
            dep_list[dep['name']]['cmp_path']  = cfg.sim_output_dir + "/" + cli_args['simulator'] + "/external/" + dep['name']
    return dep_list


def get_incdir_list(deps, cli_args):
    incdir_list = []
    for dep in deps:
        for src_path in deps[dep]['src_paths']:
            if cli_args['simulator'] == "vivado":
                incdir_list.append("-i " + src_path)
            else:
                incdir_list.append("+incdir+" + src_path)
    return incdir_list


def convert_compilation_args(cli_args):
    args = []
    return args


def convert_elaboration_args(cli_args):
    args = []
    if cli_args['simulator'] == 'vivado':
        if cli_args['waves'] or cli_args['cov']:
            args.append("--debug all")
    return args


def check_simulator_bin_path(cli_args):
    if cli_args['simulator'] == "vivado":
        if not os.path.exists(cfg.vivado_home):
            common.fatal("Path for vivado executables could not be found " + cfg.vivado_home)
    elif cli_args['simulator'] == "metrics":
        if not os.path.exists(cfg.metrics_home):
            common.fatal("Path for metrics executables could not be found " + cfg.metrics_home)
    elif cli_args['simulator'] == "vcs":
        if not os.path.exists(cfg.vcs_home):
            common.fatal("Path for vcs executables could not be found " + cfg.vcs_home)
    elif cli_args['simulator'] == "xcelium":
        if not os.path.exists(cfg.xcelium_home):
            common.fatal("Path for xcelium executables could not be found " + cfg.xcelium_home)
    elif cli_args['simulator'] == "questa":
        if not os.path.exists(cfg.questa_home):
            common.fatal("Path for questa executables could not be found " + cfg.questa_home)
    elif cli_args['simulator'] == "riviera":
        if not os.path.exists(cfg.riviera_home):
            common.fatal("Path for riviera executables could not be found " + cfg.riviera_home)
    return True


def invoke_fsoc(ip, core, full_name, ip_path, target, cli_args):
    ip_name = ip['name']
    core_name = core['name']
    cores_root = core['path']
    fsc_args = argparse.Namespace()
    fsc_args.setup       = True
    fsc_args.build       = False
    fsc_args.run         = False
    fsc_args.no_export   = True
    fsc_args.tool        = "xsim"
    fsc_args.target      = target
    fsc_args.system      = full_name
    fsc_args.backendargs = ""
    fsc_args.system_name = ""
    fsc_args.build_root  = cfg.fsoc_dir + "/" + core['sname']
    fsc_args.flag = []
    fsc_cores_root = [cores_root]
    
    try:
        fusesoc.main.init_logging(False, False)
        fsc_cfg = fusesoc.main.Config()
        fsc_cm  = fusesoc.main.init_coremanager(fsc_cfg, fsc_cores_root)
        fusesoc.main.run(fsc_cm, fsc_args)
        file_path_partial_name = re.sub(r':', '_', full_name)
        eda_file_dir = cfg.fsoc_dir + "/" + core['sname'] + "/sim-xsim"
        eda_file_path = eda_file_dir + "/" + file_path_partial_name + "_0.eda.yml"
        if not os.path.exists(eda_file_path):
            common.fatal("Could not find FuseSoC output file " + eda_file_path)
        else:
            with open(eda_file_path, 'r') as edafile:
                eda_yaml = yaml.load(edafile, Loader=SafeLoader)
                if not eda_yaml:
                    common.fatal("Could not parse FuseSoC output file " + eda_file_path)
                dirs    = []
                files   = []
                defines = []
                for file in eda_yaml['files']:
                    if file['file_type'] == "systemVerilogSource":
                        if 'is_include_file' in file:
                            if 'include_path' in file:
                                dir_path = file['include_path']
                                dir_path = dir_path.replace(core_name, "%ROOT%", 1)
                                dir_path = re.sub(".+%ROOT%", core['path'], dir_path)
                                dirs.append(dir_path)
                        else:
                            file_path = file['name']
                            file_path = file_path.replace(core_name, "%ROOT%", 1)
                            file_path = re.sub(".+%ROOT%", core['path'], file_path)
                            files.append(file_path)
                
                for param in eda_yaml['parameters']:
                    if eda_yaml['parameters'][param]['datatype'] == 'bool':
                        new_define = {};
                        #new_define['boolean'] = True
                        new_define['name'] = param
                        if eda_yaml['parameters'][param]['default'] == True:
                            new_define['value'] = 1
                        else:
                            new_define['value'] = 0
                        defines.append(new_define)
                    else:
                        common.fatal("Support for non-bool FuseSoC parameters is not currently implemented")
                sim_nickname = ""
                if cli_args['simulator'] == 'vivado':
                    sim_nickname = "xsim"
                    for option in eda_yaml['tool_options'][sim_nickname]['xelab_options']:
                        if (re.match("--define", option)):
                            new_define = {};
                            matches = re.search("--define\s+(\w+)\s*(?:=\s*(\S+))?", option)
                            if matches:
                                new_define['name'] = matches.group(1)
                                if len(matches.groups()) > 2:
                                    new_define['value'] = matches.group(2)
                                else:
                                    new_define['boolean'] = True
                                defines.append(new_define)
                else:
                    common.fatal("FuseSoC cores not yet supported for " + cli_args['simulator'])
                flist_template = cfg.templateEnv.get_template("vivado.flist.j2")
                outputText = flist_template.render(defines=defines, files=files, dirs=dirs)
                flist_file_path = str(pathlib.Path(eda_file_dir + "/" + file_path_partial_name + "_0.flist").resolve())
                with open(flist_file_path,'w') as flist_file:
                    flist_file.write(outputText)
                flist_file.close()
                with open(flist_file_path,'r') as flist_file:
                    content = flist_file.read()
                    var = "${" + core['sname'].replace("-", "_").upper() + "_SRC_PATH}"
                    content = re.sub(f"(\.\.\/)+{ip_path}/{core['sname']}", var, content, flags = re.M)
                flist_file.close()
                with open(flist_file_path,'w') as flist_file:
                    flist_file.write(content)
                flist_file.close()
                return flist_file_path
    except Exception as e:
        common.fatal("Failed to convert FuseSoC output data for core '" + core['name'] + "': "+ str(e))


def launch_eda_bin(path, args, wd, output=False):
    args_str = ""
    for arg in args:
        args_str = args_str + "  " + arg
    os.chdir(wd)
    common.dbg("Launching " + path + " with arguments '" + args_str + "' from " + wd)
    if output:
        subprocess.call(path + " " + args_str, shell=True)
    else:
        subprocess.call(path + " " + args_str + " > /dev/null 2>&1", shell=True)


def scan_cmp_log_file_for_errors(log_file_path, cli_args):
    common.dbg("Scanning compilation log file " + log_file_path + " for errors")
    errors = []
    if cli_args['simulator'] == "vivado":
        regexes = vivado_cmp_log_error_regexes
    elif cli_args['simulator'] == "vcs":
        regexes = vcs_cmp_log_error_regexes
    elif cli_args['simulator'] == "metrics":
        regexes = metrics_cmp_log_error_regexes
    elif cli_args['simulator'] == "xcelium":
        regexes = xcelium_cmp_log_error_regexes
    elif cli_args['simulator'] == "questa":
        regexes = questa_cmp_log_error_regexes
    elif cli_args['simulator'] == "riviera":
        regexes = riviera_cmp_log_error_regexes
    try:
        for i, line in enumerate(open(log_file_path)):
            for regex in regexes:
                matches = re.search(regex, line)
                if matches:
                    errors.append(line.replace("\n", ""))
    except Exception as e:
        common.fatal("Failed while parsing compilation log file " + log_file_path + ": " + str(e))
    return errors


def scan_elab_log_file_for_errors(log_file_path, cli_args):
    common.dbg("Scanning elaboration log file " + log_file_path + " for errors")
    errors = []
    if cli_args['simulator'] == "vivado":
        regexes = vivado_elab_log_error_regexes
    elif cli_args['simulator'] == "vcs":
        regexes = vcs_elab_log_error_regexes
    elif cli_args['simulator'] == "metrics":
        regexes = metrics_elab_log_error_regexes
    elif cli_args['simulator'] == "xcelium":
        regexes = xcelium_elab_log_error_regexes
    elif cli_args['simulator'] == "questa":
        regexes = questa_elab_log_error_regexes
    elif cli_args['simulator'] == "riviera":
        regexes = riviera_elab_log_error_regexes
    try:
        for i, line in enumerate(open(log_file_path)):
            for regex in regexes:
                matches = re.search(regex, line)
                if matches:
                    errors.append(line.replace("\n", ""))
    except Exception as e:
        common.fatal("Failed while parsing elaboration log file " + log_file_path + ": " + str(e))
    return errors


def log_cmp_history_fsoc(core, log_path, cli_args, timestamp_start, timestamp_end):
    core_name = core['name']
    common.dbg("Updating history with FuseSoC core '" + core_name + "' compilation")
    if core_name not in cfg.job_history:
        cfg.job_history[core_name] = {}
    if 'compilation' not in cfg.job_history[core_name]:
        cfg.job_history[core_name]['compilation'] = []
    cfg.job_history[core_name]['compilation'].append({
        "simulator"       : cli_args['simulator'],
        'timestamp_start' : timestamp_start,
        'timestamp_end'   : timestamp_end,
        'log_path'        : log_path
    })


def log_cmp_history_ip(ip_yml, log_path, cli_args, timestamp_start, timestamp_end):
    ip_name = ip_yml['name']
    common.dbg("Updating history with IP '" + ip_name + "' compilation")
    if ip_name not in cfg.job_history:
        cfg.job_history[ip_name] = {}
    if 'compilation' not in cfg.job_history[ip_name]:
        cfg.job_history[ip_name]['compilation'] = []
    cfg.job_history[ip_name]['compilation'].append({
        "simulator"       : cli_args['simulator'],
        'timestamp_start' : timestamp_start,
        'timestamp_end'   : timestamp_end,
        'log_path'        : log_path
    })


def log_cmp_history_vivado_project(ip_yml, log_path_vlog, log_path_vhdl, cli_args, timestamp_start, timestamp_end):
    ip_name = ip_yml['name']
    common.dbg("Updating history with Vivado Project IP '" + ip_name + "' compilation")
    if ip_name not in cfg.job_history:
        cfg.job_history[ip_name] = {}
    if 'compilation' not in cfg.job_history[ip_name]:
        cfg.job_history[ip_name]['compilation'] = []
    cfg.job_history[ip_name]['compilation'].append({
        "simulator"       : cli_args['simulator'],
        'timestamp_start' : timestamp_start,
        'timestamp_end'   : timestamp_end,
        'log_path_vlog'   : log_path_vlog,
        'log_path_vhdl'   : log_path_vhdl
    })


def log_elab_history(ip_yml, log_path, cli_args, timestamp_start, timestamp_end, is_regression=False, regression_name="", regression_timestamp=""):
    ip_name = ip_yml['name']
    common.dbg("Updating history with IP '" + ip_yml['name'] + "' elaboration")
    if ip_name not in cfg.job_history:
        cfg.job_history[ip_name] = {}
    if 'elaboration' not in cfg.job_history[ip_name]:
        cfg.job_history[ip_name]['elaboration'] = []
    cfg.job_history[ip_name]['elaboration'].append({
        "simulator"            : cli_args['simulator'],
        'timestamp_start'      : timestamp_start,
        'timestamp_end'        : timestamp_end,
        'log_path'             : log_path,
        "is_regression"        : is_regression,
        "regression_name"      : regression_name,
        "regression_timestamp" : regression_timestamp
    })


def log_sim_start_history(ip_yml, cli_args, timestamp, is_regression=False, regression_name="", regression_timestamp=""):
    ip_name = ip_yml['name']
    common.dbg("Updating history with IP '" + ip_name + "' simulation start")
    if ip_name not in cfg.job_history:
        cfg.job_history[ip_name] = {}
    if 'simulation' not in cfg.job_history[ip_name]:
        cfg.job_history[ip_name]['simulation'] = []
    cfg.job_history[ip_name]['simulation'].append({
        "type"                 : "start",
        "simulator"            : cli_args['simulator'],
        'timestamp'            : timestamp,
        'test_name'            : cli_args['test_name'],
        'seed'                 : cli_args['seed'],
        'waves'                : cli_args['waves'],
        'cov'                  : cli_args['cov'],
        'gui'                  : cli_args['gui'],
        "args"                 : plus_args_to_str(cli_args),
        "is_regression"        : is_regression,
        "regression_name"      : regression_name,
        "regression_timestamp" : regression_timestamp
    })


def log_sim_end_history(ip_yml, log_path, results_dir, cli_args, timestamp_start, timestamp_end, is_regression=False, regression_name="", regression_timestamp=""):
    ip_name = ip_yml['name']
    common.dbg("Updating history with IP '" + ip_name + "' simulation end")
    if ip_name not in cfg.job_history:
        cfg.job_history[ip_name] = {}
    if 'simulation' not in cfg.job_history[ip_name]:
        cfg.job_history[ip_name]['simulation'] = []
    common.dbg(f"{str(len(cfg.job_history[ip_name]['simulation']))} job history items before append()")
    entry = {
        "type"                 : "end",
        "simulator"            : cli_args['simulator'],
        'timestamp_start'      : timestamp_start,
        'timestamp_end'        : timestamp_end,
        'log_path'             : log_path,
        'test_name'            : cli_args['test_name'],
        'seed'                 : cli_args['seed'],
        'waves'                : cli_args['waves'],
        'cov'                  : cli_args['cov'],
        'gui'                  : cli_args['gui'],
        'path'                 : results_dir,
        "args"                 : plus_args_to_str(cli_args),
        "is_regression"        : is_regression,
        "regression_name"      : regression_name,
        "regression_timestamp" : regression_timestamp
    }
    cfg.job_history[ip_name]['simulation'].append(entry)
    common.dbg(f"{str(len(cfg.job_history[ip_name]['simulation']))} job history items after append()")


def plus_args_to_str(cli_args):
    str = ""
    if cli_args['args'] != None:
        for arg in cli_args['args']:
            str += f" {arg}"
    return str


def plus_args_list_to_str_list(plus_args):
    str_list = []
    for arg in plus_args:
        if plus_args[arg] == "":
            str_list.append(f"+{arg}")
        else:
            str_list.append(f"+{arg}={plus_args[arg]}")
    return str_list

