# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import common
from mio import cfg
from mio import cache

import yaml
from yaml import SafeLoader

import requests
import getpass
import tarfile
import json
from pathlib import Path
from base64 import b64decode
import os
import shutil
from alive_progress import alive_bar


base_url          = "https://mooreio.com"#"http://localhost:8080"
jwt_endpoint      = base_url + "/api/authenticate"
ips_endpoint      = base_url + "/api/ips"
versions_endpoint = base_url + "/api/versions"
version_endpoint  = base_url + "/api/version/"
licenses_endpoint = base_url + "/api/licenses"
headers = {}


def install_mio_ip(name, location):
    common.dbg(f"Installing IP '{name}' under '{location}'")
    query = {'name': name}
    response = requests.get(ips_endpoint, params=query, headers=headers)
    json_data = response.json()
    payload = None
    found_payload = False
    for ip in json_data:
        if ip['name'] == name:
            ip_id = ip['id']
            common.dbg("Found IP! name=" + name + " id=" + str(ip_id))
            license_type = ip['licenseType']
            if license_type == "PUBLIC_OPEN_SOURCE":
                versions = requests.get(versions_endpoint, headers=headers).json()
                for version in versions:
                    if version['ip']['id'] == ip_id:
                        common.dbg("Found IP version on server: " + name + " v" + version['semver'])
                        payload = version['publicPayload']
                        found_payload = True
                        break
            if license_type == "PUBLIC_ENCRYPTED":
                licenses = requests.get(licenses_endpoint, headers=headers).json()
                for license in licenses:
                    if license['targetIp']['id'] == ip_id:
                        common.dbg("Found IP license on server: " + name)
                        payload = license['payload']
                        found_payload = True
                        break
    if found_payload:
        filename = Path(location + "/" + name + '.tgz')
        filename.write_bytes(b64decode(payload))
        tar = tarfile.open(filename, "r:gz")
        ip_destination_path = location + "/" + name
        if os.path.exists(ip_destination_path):
            shutil.rmtree(ip_destination_path)
        os.mkdir(ip_destination_path)
        tar.extractall(ip_destination_path)
        tar.close()
        os.remove(filename)
    else:
        common.fatal("Could not find IP '" + name + "' on Moore.io IP Marketplace")



def login(username, password):
    global headers
    if username == None or username == "":
        username = common.prompt("Please enter your Moore.io account username:")
    if password == None or password == "":
        password = getpass.getpass(prompt='[mio] Password: ')
    payload = {
        "username": username,
        "password": password,
        "rememberMe": "true"
    }
    jwt_token_response= requests.post(jwt_endpoint, json=payload)
    #print(jwt_token_response.json())
    #print(jwt_token_response.json()['id_token'])
    headers={'Authorization':'Bearer '+jwt_token_response.json()['id_token']}


def get_deps_for_target(ip_name):
    dependencies = []
    ip_metadata = common.get_ip(ip_name)['data']
    if 'dependencies' in ip_metadata:
        for dependency in ip_metadata['dependencies']:
            dependencies.append(dependency['name'])
    else:
        common.fatal("No dependencies section in " + ip_name + "'s ip.yml")
    return dependencies


def install_mio_ips(name, ip_list, location, username, password):
    num_deps = len(ip_list)
    if ip_list == None:
        common.warning("No dependencies were found in " + name + "'s ip.yml")
    else:
        login(username, password)
        for ip in ip_list:
            if ip['name'] in cfg.ip_local_cache:
                num_deps = num_deps - 1
        common.info(f"Installing {num_deps} dependencies")
        with alive_bar(num_deps, bar = 'smooth') as bar:
            for ip in ip_list:
                if ip['name'] not in cfg.ip_local_cache:
                    bar.text(ip['name'])
                    install_mio_ip(ip['name'], location)
                    bar()


def install_all_ips_for_target(global_install, username="", password=""):
    target_name = cfg.target_ip_name
    ip_data = cache.get_ip(target_name, True)
    if global_install:
        location = cfg.user_global_ips_path
    else:
        location = cfg.dependencies_path
    
    if ip_data != None:
        cache.check_target_ip()
        deps = cfg.target_ip['data']['dependencies']
        install_mio_ips(target_name, deps, location, username, password)
        cfg.target_ip['installed'] = True
    else:
        login(username, password)
        common.banner("Installing IP " + target_name)
        install_mio_ip(target_name, location)
