# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import common
from mio import cfg
from mio import cache

import yaml
from yaml import SafeLoader

import requests
import getpass
import tarfile
import json
from pathlib import Path
from base64 import b64encode
import os
import shutil
from alive_progress import alive_bar


base_url          = "https://mooreio.com"#"http://localhost:8080"
jwt_endpoint      = base_url + "/api/authenticate"
ips_endpoint      = base_url + "/api/ips"
versions_endpoint = base_url + "/api/versions"
licenses_endpoint = base_url + "/api/licenses"
headers = {}


def publish_mio_ip(token, name):
    location = cfg.ip_local_cache[name]['path']
    common.dbg(f"Publishing IP '{name}' under '{location}'")
    query = {'name': name}
    auth_headers = {
        'Authorization'    : f"Bearer {token}",
        'Accept-Encoding'  : 'gzip, deflate, br',
        'Content-Type'     : 'application/json; charset="UTF-8"'
    }
    
    response = requests.get(ips_endpoint + "?page=0&size=1000000", params=query, headers=auth_headers)
    json_data = response.json()
    payload = None
    found_ip = False
    
    for ip in json_data:
        if ip['name'] == name:
            ip_id = ip['id']
            common.dbg("Found IP! name=" + name + " id=" + str(ip_id))
            license_type = ip['licenseType']
            if license_type == "PUBLIC_OPEN_SOURCE":
                versions = requests.get(versions_endpoint + "?page=0&size=1000000", headers=auth_headers).json()
                for version in versions:
                    if version['ip']['id'] == ip_id:
                        common.dbg("Found IP version on server: " + name + " v" + version['semver'])
                        found_ip = True
                        tarball_path = package_ip(name, location)
                        try:
                            with open(tarball_path,'rb') as f:
                                payload = b64encode(f.read())
                        except Exception as e:
                            common.fatal(f"Could not open '{tarball_path}': {e}")
                        url = versions_endpoint + f"/{version['id']}"
                        data = {
                            "id"            : version['id']     ,
                            "publicPayload" : str(payload)[2:-1]
                        }
                        common.dbg(f"Creating PATCH request to '{url}' with data '{str(data)}'")
                        try:
                            response = requests.patch(url=url, headers=auth_headers, json=data, verify=False)
                        except Exception as e:
                            common.fatal(f"Failed to publish IP '{name}': {e}")
                        common.dbg(str(response.content))
                        break
            if license_type == "PUBLIC_ENCRYPTED":
                licenses = requests.get(licenses_endpoint + "?page=0&size=1000000", headers=auth_headers).json()
                for license in licenses:
                    if license['targetIp']['id'] == ip_id:
                        common.dbg("Found IP license on server: " + name)
                        found_ip = True
                        tarball_path = package_ip(name, location)
                        try:
                            with open(tarball_path,'rb') as f:
                                payload = b64encode(f.read())
                        except Exception as e:
                            common.fatal(f"Could not open '{tarball_path}': {e}")
                        url = licenses_endpoint + f"/{license['id']}"
                        data = {
                            "id"      : license['id']     ,
                            "payload" : str(payload)[2:-1]
                        }
                        common.dbg(f"Creating PATCH request to '{url}' with data '{str(data)}'")
                        try:
                            response = requests.patch(url=url, headers=auth_headers, json=data, verify=False)
                        except Exception as e:
                            common.fatal(f"Failed to publish IP '{name}': {e}")
                        common.dbg(str(response.content))
                        break
    if found_ip:
        os.remove(tarball_path)
        common.info(f"Published IP '{name}' successfully")
    else:
        common.fatal(f"Could not find IP '{name}' on Moore.io IP Marketplace")


def package_ip(ip_name, location):
    tarball_path = cfg.temp_path + "/" + ip_name + ".tgz"
    ip_data = cache.get_ip(ip_name)['data']
    scripts_src_dir  = ip_data['structure']['scripts-path']
    docs_src_dir     = ip_data['structure']['docs-path']
    examples_src_dir = ip_data['structure']['examples-path']
    src_dir          = ip_data['structure']['src-path']
    ip_src_directories = ip_data['hdl-src']['directories']
    common.dbg(f"Creating tarball '{location}' -> '{tarball_path}'")
    
    try:
        with tarfile.open(tarball_path, "w:gz") as tar:
            tar.add(location + "/" + scripts_src_dir , arcname=scripts_src_dir )
            tar.add(location + "/" + docs_src_dir    , arcname=docs_src_dir    )
            tar.add(location + "/" + examples_src_dir, arcname=examples_src_dir)
            tar.add(location + "/" + src_dir         , arcname=src_dir         )
            for path in os.listdir(location):
                file_path = os.path.join(location, path)
                if os.path.isfile(file_path):
                    tar.add(file_path, arcname=path)
    except Exception as e:
        common.fatal(f"Failed to create tarball for IP '{ip_name}': {e}")
    return tarball_path


def login(username, password):
    global headers
    if username == None or username == "":
        username = common.prompt("Please enter your Moore.io account username:")
    if password == None or password == "":
        password = getpass.getpass(prompt='[mio] Password: ')
    payload = {
        "username": username,
        "password": password,
        "rememberMe": "true"
    }
    jwt_token_response = requests.post(jwt_endpoint, json=payload)
    token = jwt_token_response.json()['id_token']
    return token


def publish_ip(username="", password=""):
    target_name = cfg.target_ip_name
    if target_name not in cfg.ip_local_cache:
        common.fatal(f"IP '{target_name}' is not local.")
    token = login(username, password)
    publish_mio_ip(token, target_name)
