# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import common
from mio import sim
from mio import eal
from mio import cache
from mio import cfg
from mio import results
from mio import cli

import yaml
from yaml.loader import SafeLoader
from alive_progress import alive_bar
from multiprocessing import Process
import threading
from threading import BoundedSemaphore
from threading import Thread
import os
import random
import math
import time
from datetime import datetime


bar = None
sem = None
sem_cfg = None


class TestSuite:
    """TestSuite model"""

    def __init__(self, filepath, simulator):
        self.timestamp         = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        self.filepath          = filepath
        self.name              = ""
        self.ip_name           = ""
        self.cmp_args          = []
        self.elab_args         = []
        self.verbosity         = []
        self.sets              = {}
        self.regr              = {}
        self.waves             = []
        self.cov               = []
        self.max_durations     = []
        self.max_jobs          = []
        if simulator == "viv":
            self.simulator = "vivado"
        elif simulator == "mtr":
            self.simulator = "metrics"
        elif simulator == "vcs":
            self.simulator = "vcs"
        elif simulator == "xcl":
            self.simulator = "xcelium"
        elif simulator == "qst":
            self.simulator = "questa"
        elif simulator == "riv":
            self.simulator = "riviera"
        else:
            self.simulator = cfg.default_simulator
    
    def parse_yml(self, yml):
        if 'suite' not in yml:
            common.fatal("No 'suite' entry!")
        common.dbg("Parsing suite metadata")
        
        if 'info' not in yml['suite']:
            common.fatal("No 'info' entry in 'suite'!")
        common.dbg("Parsing suite info")
        
        if 'name' not in yml['suite']['info']:
            common.fatal("No 'name' entry in 'suite'!")
        self.name = yml['suite']['info']['name'].strip()
        if not isinstance(self.name, str):
            common.fatal("'name' must be a string!")
        
        if 'ip' not in yml['suite']['info']:
            common.fatal("No 'ip' entry in 'suite'!")
        self.ip_name = yml['suite']['info']['ip'].strip().lower()
        if not isinstance(self.ip_name, str):
            common.fatal("'ip_name' must be a string!")
        
        if 'target' in yml['suite']:
            common.dbg("Parsing suite target")
            if 'compilation-args' in yml['suite']['target']:
                self.cmp_args = yml['suite']['target']['compilation-args']
                if type(self.cmp_args) is not list and self.cmp_args != None:
                    common.fatal("'compilation-args' must be a list!")
                for arg in self.cmp_args:
                    if not isinstance(arg, str):
                        common.fatal("'compilation-args' entry key is not a string")
                    if not isinstance(self.cmp_args[arg], str):
                        common.fatal("'compilation-args' entry is not a string")
            
            if 'elaboration-args' in yml['suite']['target']:
                self.elab_args = yml['suite']['target']['elaboration-args']
                if type(self.elab_args) is not list and self.elab_args != None:
                    common.fatal("'elaboration-args' must be a list!")
                for arg in self.elab_args:
                    if not isinstance(arg, str):
                        common.fatal("'elaboration-args' entry key is not a string")
                    if not isinstance(self.elab_args[arg], str):
                        common.fatal("'elaboration-args' entry is not a string")
        
        if 'settings' not in yml['suite']:
            common.fatal("No 'settings' entry in 'suite'!")
        common.dbg("Parsing suite settings")
        
        if 'waves' in yml['suite']['settings']:
            self.waves = yml['suite']['settings']['waves']
            if type(self.waves) is not list:
                common.fatal("'waves' must be a list!")
                for wave in self.waves:
                    if not isinstance(wave, str):
                        common.fatal("'waves' entry is not a string")
        
        if 'cov' in yml['suite']['settings']:
            self.cov = yml['suite']['settings']['cov']
            if type(self.cov) is not list:
                common.fatal("'cov' must be a list!")
                for cov in self.cov:
                    if not isinstance(cov, str):
                        common.fatal("'cov' entry is not a string")
        
        if 'verbosity' in yml['suite']['settings']:
            self.verbosity = yml['suite']['settings']['verbosity']
            if type(self.verbosity) is not dict:
                common.fatal("'max-duration' must be a dictionary!")
            for verb in self.verbosity:
                if not isinstance(verb, str):
                    common.fatal("'verbosity' entry key is not a string!")
                if not (self.verbosity[verb] in cli.uvm_levels):
                    common.fatal(f"'verbosity' entry '{self.verbosity[verb]}' is not valid! Choices are: none, low, medium, high or debug.")
        
        if 'max-duration' in yml['suite']['settings']:
            self.max_durations = yml['suite']['settings']['max-duration']
            if type(self.max_durations) is not dict:
                common.fatal("'max-duration' must be a dictionary!")
            for duration in self.max_durations:
                if not isinstance(duration, str):
                    common.fatal("'max-duration' entry key is not a string!")
                if not (type(self.max_durations[duration]) == int or float):
                    common.fatal("'max-duration' entry is not a number!")
        else:
            common.fatal("No 'max-duration' entry in test suite settings!")
        
        if 'max-jobs' in yml['suite']['settings']:
            self.max_jobs = yml['suite']['settings']['max-jobs']
            if type(self.max_jobs) is not dict:
                common.fatal("'max-jobs' must be a dictionary!")
            for max in self.max_jobs:
                if not isinstance(max, str):
                    common.fatal("'max-jobs' entry key is not a string!")
                if not (type(self.max_jobs[max]) == int):
                    common.fatal("'max-jobs' entry is not an integer!")
        else:
            common.fatal("No 'max-jobs' entry in test suite settings!")
        
        set_count = 0
        common.dbg("Parsing sets")
        for set in yml:
            if set == "suite":
                continue
            if type(yml[set]) is not dict:
                common.fatal(f"Test set '{set}' is not a dictionary!")
            test_set = TestSet(set, self)
            self.add_test_set(test_set)
            common.dbg(f"Parsing set '{set}'")
            for group in yml[set]:
                if type(yml[set][group]) is not dict:
                    common.fatal(f"Test group '{set}.{group}' is not a dictionary!")
                common.dbg(f"Parsing group '{set}.{group}'")
                group_obj = TestGroup(group, test_set)
                test_set.add_test_group(group_obj)
                for test in yml[set][group]:
                    if type(yml[set][group][test]) is not list:
                        common.fatal(f"Test '{set}.{group}.{test}' is not a list!")
                    for regr in yml[set][group][test]:
                        if type(regr) is not dict:
                            common.fatal(f"Test group '{set}.{group}.{test}' regression is not a dictionary!")
                        if 'regr' not in regr:
                            common.fatal(f"Regression entry missing name tag ('regr'): '{set}.{group}.{test}'")
                            if not isinstance(regr['regr'], str):
                                common.fatal(f"Regression entry name is not string: '{set}.{group}.{test}'")
                        if regr['regr'] not in self.regr:
                            regression = Regression(regr['regr'].strip(), group_obj, set, self)
                            self.add_regression(regression)
                            common.dbg(f"Added regression '{regr['regr']}' from '{set}.{group}.{test}'")
                        else:
                            regression = self.regr[regr['regr']]
                        args = []
                        if 'args' in regr:
                            if type(regr['args']) is not list:
                                common.fatal(f"Regression entry args is not list: '{set}.{group}.{test}'")
                            for arg in args:
                                if not isinstance(args[arg], str):
                                    common.fatal(f"Regression entry args entry is not string: '{set}.{group}.{test}'")
                            args = regr['args']
                        if 'count' in regr:
                            if 'seeds' in regr:
                                common.fatal(f"Cannot have both 'seeds' and 'count' entry in regression entry: '{set}.{group}.{test}'")
                            if not (type(regr['count']) == int):
                                common.fatal(f"'{str(regr['count'])}' is not an integer: '{set}.{group}.{test}'!")
                            if regr['count'] <= 0:
                                common.fatal(f"'count' value '{regr['count']}' is less than 1: '{set}.{group}.{test}'")
                            for ii in range(regr['count']):
                                seed = random.randint(1, 2147483646)
                                regression.add_test(test, seed, args)
                        elif 'seeds' in regr:
                            #if len(regr['seeds']) == 0:
                            #    common.fatal(f"'seeds' length is less than 1: '{set}.{group}.{test}'")
                            for seed in regr['seeds']:
                                if not (type(seed) == int):
                                    common.fatal(f"seed value '{str(seed)}' is not an integer: '{set}.{group}.{test}'!")
                            unique_seeds = list(dict.fromkeys(regr['seeds']))
                            for seed in unique_seeds:
                                regression.add_test(test, seed, args)
                        else:
                            common.fatal(f"Illegal regression entry: '{set}.{group}.{test}'")
        
    
    def check_consistency(self):
        if self.ip_name != cfg.target_ip_name:
            common.fatal(f"IP '{self.ip_name}' in suite does not match '{cfg.target_ip_name}' as specified on the command line!")
        if len(self.sets) == 0:
            common.fatal("Test suite does not contain any test sets!")
        test_count = 0
        for regr in self.regr:
            if (self.regr[regr].name == cfg.regression_name):
                test_count = len(self.regr[regr].tests)
        if test_count == 0:
            common.fatal("Regression describes no tests.  Please add entries and try again.")
        if test_count > 1000:
            common.fatal("Regression describes more than 1,000 tests.  Please reduce this and try again.")
        if test_count > 100:
            common.warning(f"Regression describes more than 100 tests ({test_count})")
        for wave in self.waves:
            if wave not in self.regr:
                common.fatal(f"'waves' entry '{wave} is not a regression declared in this test suite")
        for cov in self.cov:
            if cov not in self.regr:
                common.fatal(f"'cov' entry '{cov} is not a regression declared in this test suite")
        for verb in self.verbosity:
            if verb not in self.regr:
                common.fatal(f"'verbosity' entry '{verb} is not a regression declared in this test suite")
        for max in self.max_durations:
            if max not in self.regr:
                common.fatal(f"'max-duration' entry '{max} is not a regression declared in this test suite")
            self.regr[max].max_duration = self.max_durations[max]
        for max in self.max_jobs:
            if max not in self.regr:
                common.fatal(f"'max-jobs' entry '{max} is not a regression declared in this test suite")
            self.regr[max].max_jobs = self.max_jobs[max]
    
    def add_test_set(self, set):
        if set.name in self.sets:
            common.fatal(f"Test group '{group.name}' already exists within test set '{self.name}'")
        else:
            self.sets[set.name] = set
            set.suite = self
    
    def add_regression(self, regr):
        self.regr[regr.name] = regr
        regr.suite = self
        common.dbg(f"Regression '{regr.name}' added to set '{self.name}'")
    
    def get_cmp_cfg(self):
        args = {}
        args['test_name'] = "undefined" # Not needed but just being safe until sim_args is turned into a class
        args['verbosity'] = "undefined" # Not needed but just being safe until sim_args is turned into a class
        args['seed'] = 0 # Not needed but just being safe until sim_args is turned into a class
        args['cmp' ] = True
        args['elab'] = False
        args['sim' ] = False
        args['gui' ] = False
        args['simulator'] = self.simulator
        args['args'] = self.cmp_args
        if cfg.regression_name in self.waves:
            args['waves'] = True
        else:
            args['waves'] = False
        if cfg.regression_name in self.cov:
            args['cov'] = True
        else:
            args['cov'] = False
        return args
    
    def get_elab_cfg(self):
        args = {}
        args['test_name'] = "undefined" # Not needed but just being safe until sim_args is turned into a class
        args['verbosity'] = "undefined" # Not needed but just being safe until sim_args is turned into a class
        args['seed'] = 0 # Not needed but just being safe until sim_args is turned into a class
        args['cmp' ] = False
        args['elab'] = True
        args['sim' ] = False
        args['gui' ] = False
        args['simulator'] = self.simulator
        args['args'] = self.elab_args
        if cfg.regression_name in self.waves:
            args['waves'] = True
        else:
            args['waves'] = False
        if cfg.regression_name in self.cov:
            args['cov'] = True
        else:
            args['cov'] = False
        return args
    
    def get_regression(self, regr_name):
        for reg in self.regr:
            if self.regr[reg].name == regr_name:
                return self.regr[reg]
        commmon.fatal(f"Could not find regression '{regr_name}'")


class TestSet:
    """Test Set model"""

    def __init__(self, name, suite):
        self.name         = name
        self.suite        = suite
        self.waves        = False
        self.cov          = False
        self.max_duration = 1
        self.max_jobs     = 1
        self.groups       = {}
    
    def add_test_group(self, group):
        if group.name in self.groups:
            common.fatal(f"Test group '{group.name}' already exists within test set '{self.name}'")
        else:
            self.groups[group.name] = group
            group.set = self
            common.dbg(f"Test group '{group.name}' added to set '{self.name}'")


class TestGroup:
    """Test Group model"""

    def __init__(self, name, set):
        self.name  = name
        self.set   = set
        self.tests = []
    
    def add_test(self, test):
        self.tests.append(test)
        common.fatal(f"Test '{test.name}' added to group '{self.name}'")


class Regression:
    """Regression model"""

    def __init__(self, name, group, set, suite):
        self.name         = name
        self.tests        = []
        self.set          = set
        self.group        = group
        self.suite        = suite
        self.max_duration = 0
        self.max_jobs     = 1
    
    def add_test(self, name, seed, args):
        test = RegressionTest(name, self, self.group, self.set, self.suite, seed, args)
        self.tests.append(test)
        common.dbg(f"Test '{name}' added to group '{self.group.name}'")
    
    def get_tests(self):
        return self.tests
    
    def reduce(self):
        removal_list = []
        ii = 0
        for curr_test in self.tests:
            jj = 0
            for other_test in self.tests:
                if other_test == curr_test:
                    continue
                else:
                    if (jj not in removal_list):
                        if curr_test.is_equal(other_test):
                            removal_list.append(ii)
                jj += 1
            ii += 1
        if len(removal_list) > 0:
            common.warning(f"Found {len(removal_list)} redundancies in regression '{name}'")
            for index in removal_list:
                self.tests[index].remove(index)


class RegressionTest:
    """Regression Test model"""

    def __init__(self, name, regression, test_group, set, suite, seed, args):
        self.name       = name
        self.regression = regression
        self.test_group = test_group
        self.set        = set
        self.suite      = suite
        self.seed       = seed
        self.args       = args
    
    def get_sim_args(self):
        args = {}
        args['test_name'] = self.name
        args['seed'] = self.seed
        args['cmp' ] = False
        args['elab'] = False
        args['sim' ] = True
        args['gui' ] = False
        args['simulator'] = self.suite.simulator
        args['args'] = self.args
        if cfg.regression_name in self.suite.waves:
            args['waves'] = True
        else:
            args['waves'] = False
        if cfg.regression_name in self.suite.cov:
            args['cov'] = True
        else:
            args['cov'] = False
        if cfg.regression_name in self.suite.verbosity:
            args['verbosity'] = self.suite.verbosity[cfg.regression_name]
        else:
            args['verbosity'] = "medium"
        
        return args
    
    def is_equal(self, other):
        if self.name != other.name:
            return False
        if self.regression != other.regression:
            return False
        if self.test_group != other.test_group:
            return False
        if self.set != other.set:
            return False
        if self.suite != other.suite:
            return False
        if self.seed != other.seed:
            return False
        if self.args.sort() != other.args.sort():
            return False
        return True


def main():
    # TODO Automatically merge coverage and generate reports
    
    if 'dependencies' in cfg.target_ip['data']:
        num_deps = len(cfg.target_ip['data']['dependencies'])
        if (cfg.target_ip['installed'] == False) and (num_deps > 0):
            common.fatal(f"You must first install this IP's {str(num_deps)} dependencies: `mio install {cfg.target_ip_name}`")
    
    timestamp_start = datetime.now()
    test_suite = scan_target_ip_for_test_suite()
    regression = test_suite.get_regression(cfg.cli_args.regr)
    regression.reduce()
    tests      = regression.get_tests()
    prep_target_ip(test_suite)
    
    sim_cfg_list = []
    for test in tests:
        sim_cfg_list.append(test.get_sim_args())
    
    launch_sim_jobs(test_suite, regression, sim_cfg_list)
    timestamp_end = datetime.now()
    if not cfg.cli_args.dry:
        regr_results = results.do_parse_results(True, test_suite.name, regression.name, test_suite.timestamp)
        print_end_of_regression_msg(regr_results, test_suite, regression, sim_cfg_list, timestamp_start, timestamp_end)


def launch_sim_jobs(test_suite, regression, sim_cfg_list):
    global bar
    global sem
    global sem_cfg
    threads = []
    regression_name = ""
    if cfg.test_suite_name == "":
        regression_name = f"{cfg.regression_name}"
        common.banner(f"Running regression '{cfg.regression_name}': {str(len(sim_cfg_list))} test(s) with {str(regression.max_duration)} hour(s) timeout")
    else:
        regression_name = f"{cfg.test_suite_name}.{cfg.regression_name}"
        common.banner(f"Running regression '{cfg.regression_name}' from test suite '{cfg.test_suite_name}': {str(len(sim_cfg_list))} test(s) with {str(regression.max_duration)} hour(s) timeout")
    
    common.create_dir(cfg.sim_dir + "/regressions/" + cfg.target_ip_name + "_" + regression_name)
    common.create_dir(cfg.sim_dir + "/regressions/" + cfg.target_ip_name + "_" + regression_name + "/" + test_suite.timestamp)
    sem = BoundedSemaphore(regression.max_jobs)
    sem_cfg = BoundedSemaphore(1)
    local_cfg = []
    with alive_bar(len(sim_cfg_list), bar = 'smooth') as bar:
        for sim_cfg in sim_cfg_list:
            common.dbg("Launching test threads")
            local_cfg.append(sim_cfg.copy())
            thread = Thread(target=launch_test, args=(test_suite,local_cfg[-1],))
            thread.daemon = True
            threads.append(thread)
            thread.start()
            time.sleep(10) # HACK!!!!!
        timeout = Thread(target=timeout_process, args=(regression.max_duration,))
        timeout.daemon = True
        timeout.start()
        common.dbg("Waiting for threads to end")
        for thread in threads:
            if thread.is_alive():
                thread.join()


def timeout_process(hours):
    total_minutes = 0
    common.dbg(f"Starting timeout process for {str(hours)} hour(s)")
    for n in range(math.ceil(hours*60)):
        time.sleep(60)
        total_minutes += 1
    common.fatal(f"Regression timed out after {str(hours)} hour(s)")


def prep_target_ip(test_suite):
    sim.main(test_suite.get_cmp_cfg(), True)
    if cfg.test_suite_name != "":
        regression_name = cfg.test_suite_name + "_" + cfg.regression_name
    else:
        regression_name = cfg.regression_name
    elab_cfg = test_suite.get_elab_cfg()
    sim.main(elab_cfg, True, regression_name, test_suite.timestamp)


def launch_test(test_suite, sim_cfg):
    global sem_cfg
    sem_cfg.acquire()
    local_cfg = sim_cfg.copy()
    if cfg.test_suite_name != "":
        regression_name = cfg.test_suite_name + "_" + cfg.regression_name
    else:
        regression_name = cfg.regression_name
    plus_args = sim.convert_cli_args_to_plusargs(local_cfg['args'])
    common.dbg("Waiting turn for simulating:\n" + str(local_cfg))
    sem_cfg.release()
    wait_for_turn()
    common.dbg("Done waiting. Starting thread for simulating:\n" + str(local_cfg))
    if cfg.cli_args.dry:
        common.info(f"  dry-run: test='{local_cfg['test_name']}' seed='{str(local_cfg['seed'])}' args='{str(local_cfg['args'])}' waves='{str(local_cfg['waves'])}' cov='{str(local_cfg['cov'])}'")
    else:
        eal.simulate(cfg.target_ip, plus_args, local_cfg, True, regression_name, test_suite.timestamp)
    common.dbg("Done simulating:\n" + str(local_cfg))
    bar()
    done_with_turn()


def wait_for_turn():
    global sem
    sem.acquire()


def done_with_turn():
    global sem
    sem.release()


def print_end_of_regression_msg(results, test_suite, regression, sim_cfg_list, timestamp_start, timestamp_end):
    duration = timestamp_end - timestamp_start
    hours = math.floor(duration.total_seconds() / 3600)
    minutes = math.floor((duration.total_seconds() - (hours*3600)) / 60)
    seconds = math.ceil(duration.total_seconds() - (hours*3600) - (minutes*60))
    duration_str = f"{str(hours)} hour(s), {str(minutes)} minute(s), {str(seconds)} second(s)"
    
    if cfg.test_suite_name != "":
        regression_name = cfg.test_suite_name + "_" + cfg.regression_name
    else:
        regression_name = cfg.regression_name
    
    results_path = cfg.sim_dir + "/regressions/" + cfg.target_ip_name + "_" + regression.name + "/" + test_suite.timestamp
    
    if results.passed:
        common.info("************************************************************************************************************************")
        common.info(f"  \033[32m\033[1m'{regression_name}' regression PASSED\033[0m")
        common.info("************************************************************************************************************************")
        common.info(f"  Duration: {duration_str}")
    else:
        common.info("************************************************************************************************************************")
        common.info(f"  \033[31m\033[1m'{regression_name}' regression FAILED: {str(results.num_failed_tests)} failure(s)\033[0m")
        common.info("************************************************************************************************************************")
        common.info(f"  Duration      : {duration_str}")
        common.info(f"  # tests passed: {str(results.num_passed_tests)}")
        common.info(f"  Passing rate  : {str(results.pct_passed)} %")
    
    common.info(f"")
    common.info(f"  HTML report      : firefox {results.html_report_path} &")
    common.info(f"  Jenkins XML      : cat     {results.xml_report_path}  &")
    common.info(f"  Results directory: pushd   {results_path}")
    


def scan_target_ip_for_test_suite():
    cfg.regression_name = cfg.cli_args.regr
    suite = None
    test_suite_name = ""
    if "." in cfg.regression_name:
        split_string = regr_name.split(".")
        cfg.test_suite_name = split_string[0]
        cfg.regression_name = split_string[1]
        common.dbg(f"Test suite name is '{test_suite_name}' and regression name is '{cfg.regression_name}'")
    else:
        common.dbg(f"Test suite name is empty and regression name is '{cfg.regression_name}'")
    
    if ('tests-path' not in cfg.target_ip['data']['hdl-src']):
        common.fatal(f"IP '{cfg.target_ip_name}' does not have 'hdl-src.tests-path' defined in its ip.yml")
    ip_tests_path = cfg.target_ip['path'] + "/" + cfg.target_ip['data']['structure']['src-path'] + "/" + cfg.target_ip['data']['hdl-src']['tests-path']
    for dirpath, dirnames, filenames in os.walk(ip_tests_path):
        for file in filenames:
            if cfg.test_suite_name == "":
                common.dbg(f"looking at file {file}")
                if file.endswith("ts.yml"):
                    filepath = dirpath + "/" + file
                    common.dbg(f"Found target test suite descriptor file at {filepath}")
                    suite = parse_test_suite_file(filepath)
                    break
            else:
                if file.endswith(".ts.yml"):
                    filepath = dirpath + "/" + file
                    ts_name = file.replace(".ts.yml", "")
                    ts_name = ts_name.replce(dirpath, "")
                    common.dbg(f"Found test suite '{ts_name}' descriptor file at {filepath}")
                    if ts_name == cfg.test_suite_name:
                        common.dbg(f"Found target test suite '{ts_name}' descriptor file at {filepath}")
                        suite = parse_test_suite_file(filepath)
                        break
    if suite == None:
        common.fatal("Could not find test suite")
    return suite


def parse_test_suite_file(file):
    test_suite = None
    #try:
    with open(file, 'r') as yamlfile:
        ip_yaml = yaml.load(yamlfile, Loader=SafeLoader)
        if ip_yaml:
            test_suite = TestSuite(file, cfg.cli_args.app)
            test_suite.parse_yml(ip_yaml)
            test_suite.check_consistency()
        else:
            common.fatal(f"Failed to parse test suite yml '{file}':" + str(e))
    #except Exception as e:
    #    common.fatal(f"Failed to parse test suite '{file}': " + str(e))
    return test_suite

