# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import cfg
from mio import clean
from mio import cov
from mio import dox
from mio import results
from mio import cache
from mio import common
from mio import eal
from mio import install
from alive_progress import alive_bar
from threading import Thread
import time
import os
from datetime import datetime
import yaml
from yaml.loader import SafeLoader
import math
import re


regex_define_pattern  = "\+define\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
regex_plusarg_pattern = "\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
bar = None
seconds_waited = 0


def main(sim_cfg, is_regression=False, regression_name="", regression_timestamp=""):
    ip_name = cfg.target_ip_name
    
    if not eal.check_simulator_bin_path(sim_cfg):
        common.fatal("Simulator '" + sim_cfg['simulator'] + "' not installed properly or environment variable missing")
    
    defines   = convert_cli_args_to_defines (sim_cfg['args'])
    plus_args = convert_cli_args_to_plusargs(sim_cfg['args'])
    create_sim_directories()
    
    if sim_cfg['cmp']:
        if cfg.target_ip['installed'] == False:
            install_deps_str = ""
            num_deps = len(cfg.target_ip['data']['dependencies'])
            while (install_deps_str != "y") and (install_deps_str != "n"):
                install_deps_str = common.prompt(str(num_deps) + " IP dependencies must first be installed.  Would you like to do so now? [y/n]").strip().lower()
            if install_deps_str == "y":
                local_str = ""
                while (local_str != "y") and (local_str != "n"):
                    local_str = common.prompt("Local install (vs. global)? [y/n]").strip().lower()
                if local_str == "y":
                    global_install = False
                else:
                    global_install = True
                common.info("Installing " + str(num_deps) + " dependencies ...")
                try:
                    install.install_all_ips_for_target(global_install)
                except Exception as e:
                    common.fatal("Error during '" + ip_name + "' IP installation: " + str(e))
                cache.scan_and_load_ip_metadata()
            else:
                common.fatal("Cannot continue without first installing '" + ip_name + "' IP dependencies.")
    
    check_dependencies()
    
    if sim_cfg['cmp']:
        if not is_regression:
            common.banner("Compiling IP '" + ip_name + "'")
        num_deps = len(cfg.target_ip['data']['dependencies'])
        for dep in cfg.target_ip['data']['dependencies']:
            if dep['name'] not in cfg.ip_local_cache:
                dep_ip = cache.get_ip(dep['name'])
                if dep_ip['compiled'][sim_cfg['simulator']]:
                    num_deps -= 1
        if num_deps > 0:
            common.info("Compiling " + str(num_deps) + " dependencies")
            cmp_dependencies(num_deps, defines, sim_cfg)
        if 'dut' in cfg.target_ip['data']:
            dut_ip_name = cfg.target_ip['data']['dut']['name']
            if dut_ip_name in cfg.job_history:
                est_time = 0
                curr_est_time = 0
                if 'compilation' in cfg.job_history[ip_name]:
                    for job in cfg.job_history[ip_name]['compilation']:
                        start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                        end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                        curr_est_time = end - start
                        curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                        est_time += curr_est_time
                    est_time = math.ceil(est_time / len(cfg.job_history[ip_name]['compilation']))
                    if est_time > 0:
                        t = Thread(target=progress_bar, args=(est_time,))
                        t.start()
                        cmp_dut(defines, sim_cfg)
                        if bar != None:
                            if (seconds_waited < est_time):
                                bar(est_time - seconds_waited)
                        t.join()
                    else:
                        cmp_dut(defines, sim_cfg)
                else:
                    cmp_dut(defines, sim_cfg)
            else:
                cmp_dut(defines, sim_cfg)
        
        if ip_name in cfg.job_history:
            if 'compilation' in cfg.job_history[ip_name]:
                est_time = 0
                curr_est_time = 0
                for job in cfg.job_history[ip_name]['compilation']:
                    start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                    end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                    curr_est_time = end - start
                    curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                    est_time += curr_est_time
                est_time = math.ceil(est_time / len(cfg.job_history[ip_name]['compilation']))
                if est_time > 0:
                    t = Thread(target=progress_bar, args=(est_time,))
                    t.start()
                    cmp_target_ip(defines, sim_cfg)
                    if bar != None:
                        if (seconds_waited < est_time):
                            bar(est_time - seconds_waited)
                    t.join()
                else:
                    cmp_target_ip(defines, sim_cfg)
            else:
                cmp_target_ip(defines, sim_cfg)
        else:
            cmp_target_ip(defines, sim_cfg)
    
    if sim_cfg['elab']:
        if is_regression:
            common.info("Elaborating IP '" + ip_name + "'")
        else:
            common.banner("Elaborating IP '" + ip_name + "'")
        est_time = 0
        curr_est_time = 0
        if ip_name in cfg.job_history:
            if 'elaboration' in cfg.job_history[ip_name]:
                for job in cfg.job_history[ip_name]['elaboration']:
                    start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                    end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                    curr_est_time = end - start
                    curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                    est_time += curr_est_time
                est_time = math.ceil(est_time / len(cfg.job_history[ip_name]['elaboration']))
                if est_time > 0:
                    t = Thread(target=progress_bar, args=(est_time,))
                    t.start()
                    eal.elaborate(cfg.target_ip, defines, sim_cfg, is_regression, regression_name, regression_timestamp)
                    if bar != None:
                        if (seconds_waited < est_time):
                            bar(est_time - seconds_waited)
                    t.join()
                else:
                    eal.elaborate(cfg.target_ip, defines, sim_cfg, is_regression, regression_name, regression_timestamp)
            else:
                eal.elaborate(cfg.target_ip, defines, sim_cfg, is_regression, regression_name, regression_timestamp)
        else:
            eal.elaborate(cfg.target_ip, defines, sim_cfg, is_regression, regression_name, regression_timestamp)
    else:
        if sim_cfg['sim'] == False and not is_regression:
            print_end_of_compilation_message(sim_cfg)
    
    if sim_cfg['sim']:
        if not is_regression:
            common.banner("Simulating IP '" + ip_name + "'")
        sim_cfg = eal.simulate(cfg.target_ip, plus_args, sim_cfg, is_regression=False)
        if sim_cfg['cmp'] == True and not is_regression:
            print_end_of_compilation_message(sim_cfg)
        if sim_cfg['elab'] == True and not is_regression:
            print_end_of_elaboration_message(sim_cfg)
        if not is_regression:
            print_end_of_simulation_message(sim_cfg)
    else:
        if sim_cfg['elab'] == True and not is_regression:
            if sim_cfg['cmp'] == True:
                print_end_of_compilation_message(sim_cfg)
            print_end_of_elaboration_message(sim_cfg)


def progress_bar(seconds):
    global bar
    global seconds_waited
    tenth_seconds_waited = 0
    with alive_bar(seconds*10, bar = 'smooth', stats="{eta} estimated", monitor=False, elapsed=True) as bar:
        bar.text("estimated")
        for x in range(seconds*10):
            time.sleep(0.1)
            tenth_seconds_waited += 1
            if (tenth_seconds_waited == 10):
                tenth_seconds_waited = 0
                seconds_waited += 1
            bar()


def convert_cli_args_to_defines(args):
    defines = {}
    if args != None:
        if type(args) is list:
            for item in args:
                regex = "\+define\+(\w+)(?:\=(\w+))?"
                result = re.match(regex, item)
                if result:
                    if len(result.groups()) > 1:
                        arg_name  = result.group(1)
                        arg_value = result.group(2)
                        if arg_value == None:
                            arg_value = ""
                    elif len(result.groups()) == 1:
                        arg_name  = result.group(1)
                        arg_value = ""
                    else:
                        continue
                else:
                    continue
                defines[arg_name] = arg_value
                common.dbg("Added define '" + arg_name + "' with value '" + arg_value + "' to list")
        else:
            all_args = re.sub("\"", "", args)
            for arg in all_args:
                result = re.match(regex_define_pattern, arg)
                if result:
                    define_name = result.group(1)
                    if len(result.groups()) > 1:
                        define_value = result.group(2)
                        if define_value == None:
                            arg_value = ""
                    else:
                        define_value = ""
                    defines[define_name] = define_value
                    common.dbg("Added define '" + define_name + "' with value '" + define_value + "' to list")
    return defines


def convert_cli_args_to_plusargs(args):
    define_regex = "\+define\+(\w+)(?:\=(\w+))?"
    plus_args = {}
    if args != None:
        if type(args) is list:
            for item in args:
                result = re.match(define_regex, item)
                if result:
                    continue
                regex = "\+(\w+)(?:\=(\w+))?"
                result = re.match(regex, item)
                if result:
                    if len(result.groups()) > 1:
                        arg_name  = result.group(1)
                        arg_value = result.group(2)
                        if arg_value == None:
                            arg_value = ""
                    elif len(result.groups()) == 1:
                        arg_name  = result.group(1)
                        arg_value = ""
                    else:
                        continue
                else:
                    continue
                plus_args[arg_name] = arg_value
                common.dbg("Added plus arg '" + arg_name + "' with value '" + arg_value + "' to list")
        else:
            all_args = re.sub("\"", "", args)
            for arg in all_args:
                result = re.match(regex_define_pattern, arg)
                if not result:
                    result = re.match(regex_plusarg_pattern, arg)
                    if result:
                        arg_name = result.group(1)
                        if len(result.groups()) > 1:
                            arg_value = result.group(2)
                            if arg_value == None:
                                arg_value = ""
                        else:
                            arg_value = ""
                        plus_args[arg_name] = arg_value
                        common.dbg("Added plus arg '" + arg_name + "' with value '" + arg_value + "' to list")
    return plus_args


def create_sim_directories():
    common.dbg("Creating sim directories")
    common.create_dir(cfg.sim_dir)
    common.create_dir(cfg.sim_output_dir)
    common.create_dir(cfg.sim_output_dir + "/vivado"             )
    common.create_dir(cfg.sim_output_dir + "/vivado/external"    )
    common.create_dir(cfg.sim_output_dir + "/vivado/local"       )
    common.create_dir(cfg.sim_output_dir + "/vivado/regressions" )
    common.create_dir(cfg.sim_output_dir + "/vivado/sim"         )
    common.create_dir(cfg.sim_output_dir + "/metrics"            )
    common.create_dir(cfg.sim_output_dir + "/metrics/external"   )
    common.create_dir(cfg.sim_output_dir + "/metrics/local"      )
    common.create_dir(cfg.sim_output_dir + "/metrics/regressions")
    common.create_dir(cfg.sim_output_dir + "/metrics/sim"        )
    common.create_dir(cfg.sim_output_dir + "/vcs"                )
    common.create_dir(cfg.sim_output_dir + "/vcs/external"       )
    common.create_dir(cfg.sim_output_dir + "/vcs/local"          )
    common.create_dir(cfg.sim_output_dir + "/vcs/regressions"    )
    common.create_dir(cfg.sim_output_dir + "/vcs/sim"            )
    common.create_dir(cfg.sim_output_dir + "/xcelium"            )
    common.create_dir(cfg.sim_output_dir + "/xcelium/external"   )
    common.create_dir(cfg.sim_output_dir + "/xcelium/local"      )
    common.create_dir(cfg.sim_output_dir + "/xcelium/regressions")
    common.create_dir(cfg.sim_output_dir + "/xcelium/sim"        )
    common.create_dir(cfg.sim_output_dir + "/questa"             )
    common.create_dir(cfg.sim_output_dir + "/questa/external"    )
    common.create_dir(cfg.sim_output_dir + "/questa/local"       )
    common.create_dir(cfg.sim_output_dir + "/questa/regressions" )
    common.create_dir(cfg.sim_output_dir + "/questa/sim"         )
    common.create_dir(cfg.sim_output_dir + "/riviera"            )
    common.create_dir(cfg.sim_output_dir + "/riviera/external"   )
    common.create_dir(cfg.sim_output_dir + "/riviera/local"      )
    common.create_dir(cfg.sim_output_dir + "/riviera/regressions")
    common.create_dir(cfg.sim_output_dir + "/riviera/sim"        )
    common.create_dir(cfg.sim_dir + "/cmp")
    common.create_dir(cfg.sim_dir + "/elab")
    common.create_dir(cfg.regr_results_dir)
    common.create_dir(cfg.sim_results_dir)


def check_dependencies():
    for dep in cfg.target_ip['data']['dependencies']:
        found_ip = False
        dep_name = dep['name']
        if dep_name in cfg.ip_local_cache:
            found_ip = True
        else:
            for vendor in cfg.ip_external_cache:
                if dep_name in cfg.ip_external_cache[vendor]:
                    found_ip = True
        if not found_ip:
            common.fatal("Could not find IP dependency '" + dep_name + "'")


def cmp_dependencies(num_deps, defines, sim_cfg):
    with alive_bar(num_deps-1, bar = 'smooth') as bar:
        for dep in cfg.target_ip['data']['dependencies']:
            dep_name = dep['name']
            if dep_name == "uvm":
                continue
            if dep_name in cfg.ip_local_cache:
                dep_ip = cfg.ip_local_cache[dep_name]
                common.dbg("Compiling IP dependency '" + dep_name + "'")
                bar.text(dep_name)
                eal.compile_ip(dep_ip, defines, sim_cfg, local=True)
                bar()
            else:
                dep_ip = cache.get_external_ip(dep_name)
                if not dep_ip['compiled'][sim_cfg['simulator']]:
                    common.dbg("Compiling IP dependency '" + dep_name + "'")
                    bar.text(dep_name)
                    eal.compile_ip(dep_ip, defines, sim_cfg, local=False)
                    bar()


def cmp_dut(defines, sim_cfg):
    found_dut_ip = False
    dut_type = cfg.target_ip['data']['dut']['ip-type']
    dut_ip_name = ""
    if dut_type == "fsoc":
        dut_ip_name = cfg.target_ip['data']['dut']['full-name'].strip()
        fsoc_target = cfg.target_ip['data']['dut']['target'].strip()
        if dut_ip_name in cfg.fsoc_cache:
            found_dut_ip = True
            common.dbg("Found FuseSoC DUT '" + dut_ip_name + "'")
            core = cfg.fsoc_cache[dut_ip_name]
            ip_path = core['ip_path'].strip()
            if not core['installed'] or cfg.cli_args.F:
                flist_path = eal.invoke_fsoc(cfg.target_ip, core, dut_ip_name, ip_path, fsoc_target, sim_cfg)
                core['installed'] = True
                eal.compile_fsoc_core(flist_path, core, defines, sim_cfg)
            else:
                common.info("Skipping installation of DUT FuseSoC core '" + dut_ip_name + "'.")
    elif dut_type == "mio":
        local_ip = False
        dut_ip_name = cfg.target_ip['data']['dut']['name']
        compile_dut = False
        if dut_ip_name in cfg.ip_local_cache:
            local_ip = True
            found_dut_ip = True
            dut_ip = cfg.ip_local_cache[dut_ip_name]
            common.dbg("Found local IP DUT '" + dut_ip_name + "'")
            compile_dut = True
        else:
            dut_ip = cache.get_external_ip(dut_ip_name)
            found_dut_ip = True
            common.dbg("Found external IP DUT '" + dut_ip_name + "'")
            if not dut_ip['compiled'][sim_cfg['simulator']]:
                compile_dut = True
        if compile_dut:
            dut_sub_type = dut_ip['data']['ip']['sub-type'].lower().strip()
            if dut_sub_type == "vivado":
                common.info("Compiling Vivado Project DUT IP '" + dut_ip_name + "'")
                eal.compile_vivado_project(dut_ip, defines, sim_cfg, local=local_ip)
            elif dut_sub_type == "flist":
                common.info("Compiling DUT IP '" + dut_ip_name + "'")
                eal.compile_ip(dut_ip, defines, sim_cfg, local=local_ip)
            else:
                common.fatal("Invalid DUT sub-type '" + dut_sub_type + "'")
        else:
            common.info("Skipping compilation of DUT IP '" + dut_ip_name + "'.")
    else:
        common.fatal("DUT type '" + dut_type + "' not supported.")
    if found_dut_ip == False:
        common.fatal("Could not find DUT '" + dut_ip_name + "'.")


def cmp_target_ip(defines, sim_cfg):
    ip_name = cfg.target_ip_name
    if ip_name in cfg.ip_local_cache:
        ip_yml = cfg.ip_local_cache[ip_name]
        local_ip = True
    else:
        ip_yml = get_external_ip(ip_name)
        local_ip = False
    if 'xilinx-libs' in ip_yml['data']['hdl-src']:
        common.info("Compiling '" + ip_name + "'")
        eal.compile_vivado_project(ip_yml, defines, sim_cfg, local=local_ip)
    else:
        common.info(f"Compiling '{ip_name}'")
        eal.compile_ip(ip_yml, defines, sim_cfg, local=local_ip)



def print_end_of_compilation_message(sim_cfg):
    common.info("************************************************************************************************************************")
    common.info("  Compilation results")
    common.info("************************************************************************************************************************")
    common.info("  emacs " + cfg.sim_dir + "/cmp/" + cfg.target_ip_name + "." + common.get_simulator_short_name(sim_cfg) + ".cmp.log &")
    common.info("  gvim  " + cfg.sim_dir + "/cmp/" + cfg.target_ip_name + "." + common.get_simulator_short_name(sim_cfg) + ".cmp.log &")
    common.info("  vim   " + cfg.sim_dir + "/cmp/" + cfg.target_ip_name + "." + common.get_simulator_short_name(sim_cfg) + ".cmp.log")
    common.info("")


def print_end_of_elaboration_message(sim_cfg):
    common.info("************************************************************************************************************************")
    common.info("  Elaboration results")
    common.info("************************************************************************************************************************")
    common.info("  emacs " + cfg.sim_dir + "/elab/" + cfg.target_ip_name + "." + common.get_simulator_short_name(sim_cfg) + ".elab.log &")
    common.info("  gvim  " + cfg.sim_dir + "/elab/" + cfg.target_ip_name + "." + common.get_simulator_short_name(sim_cfg) + ".elab.log &")
    common.info("  vim   " + cfg.sim_dir + "/elab/" + cfg.target_ip_name + "." + common.get_simulator_short_name(sim_cfg) + ".elab.log")
    common.info("")


def print_end_of_simulation_message(sim_cfg):
    results_path = sim_cfg['results_path']
    common.info("************************************************************************************************************************")
    common.info("  Simulation results")
    common.info("************************************************************************************************************************")
    if (sim_cfg['waves'] == True):
        if sim_cfg['simulator'] == "vivado":
            common.info("  Waveforms: $VIVADO_HOME/xsim -gui " + results_path + "/waves.wdb &")
        elif sim_cfg['simulator'] == "vcs":
            common.info("  Waveforms: $VCS_HOME/dve -gui " + results_path + "/waves.wdb &")
        elif sim_cfg['simulator'] == "metrics":
            common.info("  Waveforms: $METRICS_HOME/??? -gui " + results_path + "/waves.wdb &")
        elif sim_cfg['simulator'] == "xcelium":
            common.info("  Waveforms: $XCELIUM_HOME/simvision -gui " + results_path + "/waves.wdb &")
        elif sim_cfg['simulator'] == "questa":
            common.info("  Waveforms: $QUESTA_HOME/visualizer -gui " + results_path + "/waves.wdb &")
        elif sim_cfg['simulator'] == "riviera":
            common.info("  Waveforms: $RIVIERA_HOME/??? -gui " + results_path + "/waves.wdb &")
    common.info("")
    common.info("  Main log: emacs " + results_path + "/sim.log &")
    common.info("            gvim  " + results_path + "/sim.log &")
    common.info("            vim   " + results_path + "/sim.log")
    common.info("")
    common.info("  Transaction logs: pushd " + results_path + "/trn_log")
    common.info("  Test result dir : pushd " + results_path)
    common.info("")


